# Copyright (C) 2006-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from ... import lazy_import

lazy_import.lazy_import(globals(), """
from breezy.bzr.smart import request as _mod_request
""")

import breezy

from ... import debug, errors, hooks, trace
from . import message, protocol


class _SmartClient:

    def __init__(self, medium, headers=None):
        """Constructor.

        :param medium: a SmartClientMedium
        """
        self._medium = medium
        if headers is None:
            self._headers = {
                b'Software version': breezy.__version__.encode('utf-8')}
        else:
            self._headers = dict(headers)

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, self._medium)

    def _call_and_read_response(self, method, args, body=None, readv_body=None,
                                body_stream=None, expect_response_body=True):
        request = _SmartClientRequest(self, method, args, body=body,
                                      readv_body=readv_body, body_stream=body_stream,
                                      expect_response_body=expect_response_body)
        return request.call_and_read_response()

    def call(self, method, *args):
        """Call a method on the remote server."""
        result, protocol = self.call_expecting_body(method, *args)
        protocol.cancel_read_body()
        return result

    def call_expecting_body(self, method, *args):
        """Call a method and return the result and the protocol object.

        The body can be read like so::

            result, smart_protocol = smart_client.call_expecting_body(...)
            body = smart_protocol.read_body_bytes()
        """
        return self._call_and_read_response(
            method, args, expect_response_body=True)

    def call_with_body_bytes(self, method, args, body):
        """Call a method on the remote server with body bytes."""
        if not isinstance(method, bytes):
            raise TypeError('method must be a byte string, not {!r}'.format(method))
        for arg in args:
            if not isinstance(arg, bytes):
                raise TypeError('args must be byte strings, not {!r}'.format(args))
        if not isinstance(body, bytes):
            raise TypeError('body must be byte string, not {!r}'.format(body))
        response, response_handler = self._call_and_read_response(
            method, args, body=body, expect_response_body=False)
        return response

    def call_with_body_bytes_expecting_body(self, method, args, body):
        """Call a method on the remote server with body bytes."""
        if not isinstance(method, bytes):
            raise TypeError('method must be a byte string, not {!r}'.format(method))
        for arg in args:
            if not isinstance(arg, bytes):
                raise TypeError('args must be byte strings, not {!r}'.format(args))
        if not isinstance(body, bytes):
            raise TypeError('body must be byte string, not {!r}'.format(body))
        response, response_handler = self._call_and_read_response(
            method, args, body=body, expect_response_body=True)
        return (response, response_handler)

    def call_with_body_readv_array(self, args, body):
        response, response_handler = self._call_and_read_response(
            args[0], args[1:], readv_body=body, expect_response_body=True)
        return (response, response_handler)

    def call_with_body_stream(self, args, stream):
        response, response_handler = self._call_and_read_response(
            args[0], args[1:], body_stream=stream,
            expect_response_body=False)
        return (response, response_handler)

    def remote_path_from_transport(self, transport):
        """Convert transport into a path suitable for using in a request.

        Note that the resulting remote path doesn't encode the host name or
        anything but path, so it is only safe to use it in requests sent over
        the medium from the matching transport.
        """
        return self._medium.remote_path_from_transport(transport).encode('utf-8')


class _SmartClientRequest:
    """Encapsulate the logic for a single request.

    This class handles things like reconnecting and sending the request a
    second time when the connection is reset in the middle. It also handles the
    multiple requests that get made if we don't know what protocol the server
    supports yet.

    Generally, you build up one of these objects, passing in the arguments that
    you want to send to the server, and then use 'call_and_read_response' to
    get the response from the server.
    """

    def __init__(self, client, method, args, body=None, readv_body=None,
                 body_stream=None, expect_response_body=True):
        self.client = client
        self.method = method
        self.args = args
        self.body = body
        self.readv_body = readv_body
        self.body_stream = body_stream
        self.expect_response_body = expect_response_body

    def call_and_read_response(self):
        """Send the request to the server, and read the initial response.

        This doesn't read all of the body content of the response, instead it
        returns (response_tuple, response_handler). response_tuple is the 'ok',
        or 'error' information, and 'response_handler' can be used to get the
        content stream out.
        """
        self._run_call_hooks()
        protocol_version = self.client._medium._protocol_version
        if protocol_version is None:
            return self._call_determining_protocol_version()
        else:
            return self._call(protocol_version)

    def _is_safe_to_send_twice(self):
        """Check if the current method is re-entrant safe."""
        if self.body_stream is not None or 'noretry' in debug.debug_flags:
            # We can't restart a body stream that has already been consumed.
            return False
        request_type = _mod_request.request_handlers.get_info(self.method)
        if request_type in ('read', 'idem', 'semi'):
            return True
        # If we have gotten this far, 'stream' cannot be retried, because we
        # already consumed the local stream.
        if request_type in ('semivfs', 'mutate', 'stream'):
            return False
        trace.mutter('Unknown request type: %s for method %s'
                     % (request_type, self.method))
        return False

    def _run_call_hooks(self):
        if not _SmartClient.hooks['call']:
            return
        params = CallHookParams(self.method, self.args, self.body,
                                self.readv_body, self.client._medium)
        for hook in _SmartClient.hooks['call']:
            hook(params)

    def _call(self, protocol_version):
        """We know the protocol version.

        So this just sends the request, and then reads the response. This is
        where the code will be to retry requests if the connection is closed.
        """
        response_handler = self._send(protocol_version)
        try:
            response_tuple = response_handler.read_response_tuple(
                expect_body=self.expect_response_body)
        except errors.ConnectionReset as e:
            self.client._medium.reset()
            if not self._is_safe_to_send_twice():
                raise
            trace.warning('ConnectionReset reading response for %r, retrying'
                          % (self.method,))
            trace.log_exception_quietly()
            encoder, response_handler = self._construct_protocol(
                protocol_version)
            self._send_no_retry(encoder)
            response_tuple = response_handler.read_response_tuple(
                expect_body=self.expect_response_body)
        return (response_tuple, response_handler)

    def _call_determining_protocol_version(self):
        """Determine what protocol the remote server supports.

        We do this by placing a request in the most recent protocol, and
        handling the UnexpectedProtocolVersionMarker from the server.
        """
        last_err = None
        for protocol_version in [3, 2]:
            if protocol_version == 2:
                # If v3 doesn't work, the remote side is older than 1.6.
                self.client._medium._remember_remote_is_before((1, 6))
            try:
                response_tuple, response_handler = self._call(protocol_version)
            except errors.UnexpectedProtocolVersionMarker as err:
                # TODO: We could recover from this without disconnecting if
                # we recognise the protocol version.
                trace.warning(
                    'Server does not understand Bazaar network protocol %d,'
                    ' reconnecting.  (Upgrade the server to avoid this.)'
                    % (protocol_version,))
                self.client._medium.disconnect()
                last_err = err
                continue
            except errors.ErrorFromSmartServer:
                # If we received an error reply from the server, then it
                # must be ok with this protocol version.
                self.client._medium._protocol_version = protocol_version
                raise
            else:
                self.client._medium._protocol_version = protocol_version
                return response_tuple, response_handler
        raise errors.SmartProtocolError(
            'Server is not a Bazaar server: ' + str(last_err))

    def _construct_protocol(self, version):
        """Build the encoding stack for a given protocol version."""
        request = self.client._medium.get_request()
        if version == 3:
            request_encoder = protocol.ProtocolThreeRequester(request)
            response_handler = message.ConventionalResponseHandler()
            response_proto = protocol.ProtocolThreeDecoder(
                response_handler, expect_version_marker=True)
            response_handler.setProtoAndMediumRequest(response_proto, request)
        elif version == 2:
            request_encoder = protocol.SmartClientRequestProtocolTwo(request)
            response_handler = request_encoder
        else:
            request_encoder = protocol.SmartClientRequestProtocolOne(request)
            response_handler = request_encoder
        return request_encoder, response_handler

    def _send(self, protocol_version):
        """Encode the request, and send it to the server.

        This will retry a request if we get a ConnectionReset while sending the
        request to the server. (Unless we have a body_stream that we have
        already started consuming, since we can't restart body_streams)

        :return: response_handler as defined by _construct_protocol
        """
        encoder, response_handler = self._construct_protocol(protocol_version)
        try:
            self._send_no_retry(encoder)
        except errors.ConnectionReset as e:
            # If we fail during the _send_no_retry phase, then we can
            # be confident that the server did not get our request, because we
            # haven't started waiting for the reply yet. So try the request
            # again. We only issue a single retry, because if the connection
            # really is down, there is no reason to loop endlessly.

            # Connection is dead, so close our end of it.
            self.client._medium.reset()
            if (('noretry' in debug.debug_flags) or
                (self.body_stream is not None and
                    encoder.body_stream_started)):
                # We can't restart a body_stream that has been partially
                # consumed, so we don't retry.
                # Note: We don't have to worry about
                #   SmartClientRequestProtocolOne or Two, because they don't
                #   support client-side body streams.
                raise
            trace.warning('ConnectionReset calling %r, retrying'
                          % (self.method,))
            trace.log_exception_quietly()
            encoder, response_handler = self._construct_protocol(
                protocol_version)
            self._send_no_retry(encoder)
        return response_handler

    def _send_no_retry(self, encoder):
        """Just encode the request and try to send it."""
        encoder.set_headers(self.client._headers)
        if self.body is not None:
            if self.readv_body is not None:
                raise AssertionError(
                    "body and readv_body are mutually exclusive.")
            if self.body_stream is not None:
                raise AssertionError(
                    "body and body_stream are mutually exclusive.")
            encoder.call_with_body_bytes(
                (self.method, ) + self.args, self.body)
        elif self.readv_body is not None:
            if self.body_stream is not None:
                raise AssertionError(
                    "readv_body and body_stream are mutually exclusive.")
            encoder.call_with_body_readv_array((self.method, ) + self.args,
                                               self.readv_body)
        elif self.body_stream is not None:
            encoder.call_with_body_stream((self.method, ) + self.args,
                                          self.body_stream)
        else:
            encoder.call(self.method, *self.args)


class SmartClientHooks(hooks.Hooks):

    def __init__(self):
        hooks.Hooks.__init__(
            self, "breezy.bzr.smart.client", "_SmartClient.hooks")
        self.add_hook('call',
                      "Called when the smart client is submitting a request to the "
                      "smart server. Called with a breezy.bzr.smart.client.CallHookParams "
                      "object. Streaming request bodies, and responses, are not "
                      "accessible.", None)


_SmartClient.hooks = SmartClientHooks()  # type: ignore


class CallHookParams:

    def __init__(self, method, args, body, readv_body, medium):
        self.method = method
        self.args = args
        self.body = body
        self.readv_body = readv_body
        self.medium = medium

    def __repr__(self):
        attrs = {k: v for k, v in self.__dict__.items()
                     if v is not None}
        return '<{} {!r}>'.format(self.__class__.__name__, attrs)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self == other



# Copyright (C) 2005-2012 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

__docformat__ = "google"

from typing import (TYPE_CHECKING, Dict, List, Optional, TextIO, Tuple, Union,
                    cast)

from .lazy_import import lazy_import

lazy_import(globals(), """
from breezy import (
    ui,
    )
from breezy.bzr import (
    fetch,
    remote,
    vf_search,
    )
""")

import contextlib
import itertools

from . import config as _mod_config
from . import debug, errors, registry, repository
from . import revision as _mod_revision
from . import urlutils
from .controldir import (ControlComponent, ControlComponentFormat,
                         ControlComponentFormatRegistry, ControlDir)
from .hooks import Hooks
from .inter import InterObject
from .lock import LogicalLockResult
from .revision import RevisionID
from .trace import is_quiet, mutter, mutter_callsite, note, warning
from .transport import Transport, get_transport

if TYPE_CHECKING:
    from .tag import TagConflict, TagUpdates



class UnstackableBranchFormat(errors.BzrError):

    _fmt = ("The branch '%(url)s'(%(format)s) is not a stackable format. "
            "You will need to upgrade the branch to permit branch stacking.")

    def __init__(self, format, url):
        errors.BzrError.__init__(self)
        self.format = format
        self.url = url


class BindingUnsupported(errors.UnsupportedOperation):

    _fmt = "Branch at %(url)s does not support binding."

    def __init__(self, branch):
        errors.BzrError.__init__(self)
        self.branch = branch
        self.url = branch.user_url


class Branch(ControlComponent):
    """Branch holding a history of revisions.

    Attributes:
      hooks: An instance of BranchHooks.
      _master_branch_cache: cached result of get_master_branch, see
        _clear_cached_state.
    """

    controldir: ControlDir

    name: Optional[str]

    base: str

    _format: "BranchFormat"

    _last_revision_info_cache: Optional[Tuple[int, RevisionID]]

    repository: repository.Repository

    hooks: "BranchHooks"

    @property
    def user_transport(self) -> Transport:
        return self.controldir.user_transport

    def __init__(self, possible_transports: Optional[List[Transport]] = None) -> None:
        self.tags = self._format.make_tags(self)
        self._revision_history_cache = None
        self._revision_id_to_revno_cache = None
        self._partial_revision_id_to_revno_cache: Dict[RevisionID, int] = {}
        self._partial_revision_history_cache: List[RevisionID] = []
        self._last_revision_info_cache = None
        self._master_branch_cache = None
        self._merge_sorted_revisions_cache = None
        self._open_hook(possible_transports)
        hooks = Branch.hooks['open']
        for hook in hooks:
            hook(self)

    def _open_hook(self, possible_transports):
        """Called by init to allow simpler extension of the base class."""

    def _activate_fallback_location(self, url, possible_transports):
        """Activate the branch/repository from url as a fallback repository."""
        for existing_fallback_repo in self.repository._fallback_repositories:
            if existing_fallback_repo.user_url == url:
                # This fallback is already configured.  This probably only
                # happens because ControlDir.sprout is a horrible mess.  To
                # avoid confusing _unstack we don't add this a second time.
                mutter('duplicate activation of fallback %r on %r', url, self)
                return
        repo = self._get_fallback_repository(url, possible_transports)
        if repo.has_same_location(self.repository):
            raise errors.UnstackableLocationError(self.user_url, url)
        self.repository.add_fallback_repository(repo)

    def break_lock(self) -> None:
        """Break a lock if one is present from another instance.

        Uses the ui factory to ask for confirmation if the lock may be from
        an active process.

        This will probe the repository for its lock as well.
        """
        raise NotImplementedError(self.break_lock)

    def _extend_partial_history(
            self, stop_index: Optional[int] = None,
            stop_revision: Optional[RevisionID] = None) -> None:
        """Extend the partial history to include a given index

        If a stop_index is supplied, stop when that index has been reached.
        If a stop_revision is supplied, stop when that revision is
        encountered.  Otherwise, stop when the beginning of history is
        reached.

        Args:
          stop_index: The index which should be present.  When it is
            present, history extension will stop.
          stop_revision: The revision id which should be present.  When
            it is encountered, history extension will stop.
        """
        if len(self._partial_revision_history_cache) == 0:
            self._partial_revision_history_cache = [self.last_revision()]
        repository._iter_for_revno(
            self.repository, self._partial_revision_history_cache,
            stop_index=stop_index, stop_revision=stop_revision)
        if self._partial_revision_history_cache[-1] == \
                _mod_revision.NULL_REVISION:
            self._partial_revision_history_cache.pop()

    def _get_check_refs(self):
        """Get the references needed for check().

        See breezy.check.
        """
        revid = self.last_revision()
        return [('revision-existence', revid), ('lefthand-distance', revid)]

    @staticmethod
    def open(base, _unsupported=False, possible_transports=None):
        """Open the branch rooted at base.

        For instance, if the branch is at URL/.bzr/branch,
        Branch.open(URL) -> a Branch instance.
        """
        control = ControlDir.open(
            base, possible_transports=possible_transports,
            _unsupported=_unsupported)
        return control.open_branch(
            unsupported=_unsupported,
            possible_transports=possible_transports)

    @staticmethod
    def open_from_transport(transport: Transport, name: Optional[str] = None,
                            _unsupported: bool = False,
                            possible_transports=None):
        """Open the branch rooted at transport"""
        control = ControlDir.open_from_transport(
            transport, _unsupported)
        return control.open_branch(
            name=name, unsupported=_unsupported,
            possible_transports=possible_transports)

    @staticmethod
    def open_containing(url, possible_transports=None):
        """Open an existing branch which contains url.

        This probes for a branch at url, and searches upwards from there.

        Basically we keep looking up until we find the control directory or
        run into the root.  If there isn't one, raises NotBranchError.
        If there is one and it is either an unrecognised format or an
        unsupported format, UnknownFormatError or UnsupportedFormatError are
        raised.  If there is one, it is returned, along with the unused portion
        of url.
        """
        control, relpath = ControlDir.open_containing(
            url, possible_transports)
        branch = control.open_branch(possible_transports=possible_transports)
        return (branch, relpath)

    def _push_should_merge_tags(self):
        """Should _basic_push merge this branch's tags into the target?

        The default implementation returns False if this branch has no tags,
        and True the rest of the time.  Subclasses may override this.
        """
        return self.supports_tags() and self.tags.get_tag_dict()

    def get_config(self):
        """Get a breezy.config.BranchConfig for this Branch.

        This can then be used to get and set configuration options for the
        branch.

        Returns: A breezy.config.BranchConfig.
        """
        return _mod_config.BranchConfig(self)

    def get_config_stack(self) -> _mod_config.Stack:
        """Get a breezy.config.BranchStack for this Branch.

        This can then be used to get and set configuration options for the
        branch.

        Returns: A breezy.config.BranchStack.
        """
        return _mod_config.BranchStack(self)

    def store_uncommitted(self, creator) -> None:
        """Store uncommitted changes from a ShelfCreator.

        Args:
          creator: The ShelfCreator containing uncommitted changes, or
            None to delete any stored changes.
        :raises: ChangesAlreadyStored if the branch already has changes.
        """
        raise NotImplementedError(self.store_uncommitted)

    def get_unshelver(self, tree):
        """Return a shelf.Unshelver for this branch and tree.

        Args:
          tree: The tree to use to construct the Unshelver.
        Returns: an Unshelver or None if no changes are stored.
        """
        raise NotImplementedError(self.get_unshelver)

    def _get_fallback_repository(self, url, possible_transports):
        """Get the repository we fallback to at url."""
        url = urlutils.join(self.base, url)
        a_branch = Branch.open(url, possible_transports=possible_transports)
        return a_branch.repository

    def _get_nick(self, local=False, possible_transports=None):
        config = self.get_config()
        # explicit overrides master, but don't look for master if local is True
        if not local and not config.has_explicit_nickname():
            try:
                master = self.get_master_branch(possible_transports)
                if master and self.user_url == master.user_url:
                    raise errors.RecursiveBind(self.user_url)
                if master is not None:
                    # return the master branch value
                    return master.nick
            except errors.RecursiveBind as e:
                raise e
            except errors.BzrError as e:
                # Silently fall back to local implicit nick if the master is
                # unavailable
                mutter("Could not connect to bound branch, "
                       "falling back to local nick.\n " + str(e))
        return config.get_nickname()

    def _set_nick(self, nick):
        self.get_config().set_user_option('nickname', nick, warn_masked=True)

    nick = property(_get_nick, _set_nick)

    def is_locked(self):
        raise NotImplementedError(self.is_locked)

    def _lefthand_history(self, revision_id, last_rev=None,
                          other_branch=None):
        if 'evil' in debug.debug_flags:
            mutter_callsite(4, "_lefthand_history scales with history.")
        # stop_revision must be a descendant of last_revision
        graph = self.repository.get_graph()
        if last_rev is not None:
            if not graph.is_ancestor(last_rev, revision_id):
                # our previous tip is not merged into stop_revision
                raise errors.DivergedBranches(self, other_branch)
        # make a new revision history from the graph
        parents_map = graph.get_parent_map([revision_id])
        if revision_id not in parents_map:
            raise errors.NoSuchRevision(self, revision_id)
        current_rev_id = revision_id
        new_history = []
        check_not_reserved_id = _mod_revision.check_not_reserved_id
        # Do not include ghosts or graph origin in revision_history
        while (current_rev_id in parents_map
               and len(parents_map[current_rev_id]) > 0):
            check_not_reserved_id(current_rev_id)
            new_history.append(current_rev_id)
            current_rev_id = parents_map[current_rev_id][0]
            parents_map = graph.get_parent_map([current_rev_id])
        new_history.reverse()
        return new_history

    def lock_write(self, token=None):
        """Lock the branch for write operations.

          token: A token to permit reacquiring a previously held and
            preserved lock.
        Returns: A BranchWriteLockResult.
        """
        raise NotImplementedError(self.lock_write)

    def lock_read(self):
        """Lock the branch for read operations.

        Returns: A breezy.lock.LogicalLockResult.
        """
        raise NotImplementedError(self.lock_read)

    def unlock(self):
        raise NotImplementedError(self.unlock)

    def peek_lock_mode(self):
        """Return lock mode for the Branch: 'r', 'w' or None"""
        raise NotImplementedError(self.peek_lock_mode)

    def get_physical_lock_status(self):
        raise NotImplementedError(self.get_physical_lock_status)

    def dotted_revno_to_revision_id(self, revno, _cache_reverse=False):
        """Return the revision_id for a dotted revno.

        Args:
          revno: a tuple like (1,) or (1,1,2)
          _cache_reverse: a private parameter enabling storage
           of the reverse mapping in a top level cache. (This should
           only be done in selective circumstances as we want to
           avoid having the mapping cached multiple times.)
        Returns: the revision_id
        :raises errors.NoSuchRevision: if the revno doesn't exist
        """
        with self.lock_read():
            rev_id = self._do_dotted_revno_to_revision_id(revno)
            if _cache_reverse:
                self._partial_revision_id_to_revno_cache[rev_id] = revno
            return rev_id

    def _do_dotted_revno_to_revision_id(self, revno):
        """Worker function for dotted_revno_to_revision_id.

        Subclasses should override this if they wish to
        provide a more efficient implementation.
        """
        if len(revno) == 1:
            try:
                return self.get_rev_id(revno[0])
            except errors.RevisionNotPresent as exc:
                raise errors.GhostRevisionsHaveNoRevno(revno[0], exc.revision_id) from exc
        revision_id_to_revno = self.get_revision_id_to_revno_map()
        revision_ids = [revision_id for revision_id, this_revno
                        in revision_id_to_revno.items()
                        if revno == this_revno]
        if len(revision_ids) == 1:
            return revision_ids[0]
        else:
            revno_str = '.'.join(map(str, revno))
            raise errors.NoSuchRevision(self, revno_str)

    def revision_id_to_dotted_revno(self, revision_id):
        """Given a revision id, return its dotted revno.

        Returns: a tuple like (1,) or (400,1,3).
        """
        with self.lock_read():
            return self._do_revision_id_to_dotted_revno(revision_id)

    def _do_revision_id_to_dotted_revno(self, revision_id):
        """Worker function for revision_id_to_revno."""
        # Try the caches if they are loaded
        result = self._partial_revision_id_to_revno_cache.get(revision_id)
        if result is not None:
            return result
        if self._revision_id_to_revno_cache:
            result = self._revision_id_to_revno_cache.get(revision_id)
            if result is None:
                raise errors.NoSuchRevision(self, revision_id)
        # Try the mainline as it's optimised
        try:
            revno = self.revision_id_to_revno(revision_id)
            return (revno,)
        except errors.NoSuchRevision as exc:
            # We need to load and use the full revno map after all
            result = self.get_revision_id_to_revno_map().get(revision_id)
            if result is None:
                raise errors.NoSuchRevision(self, revision_id) from exc
        return result

    def get_revision_id_to_revno_map(self):
        """Return the revision_id => dotted revno map.

        This will be regenerated on demand, but will be cached.

        Returns: A dictionary mapping revision_id => dotted revno.
            This dictionary should not be modified by the caller.
        """
        if 'evil' in debug.debug_flags:
            mutter_callsite(
                3, "get_revision_id_to_revno_map scales with ancestry.")
        with self.lock_read():
            if self._revision_id_to_revno_cache is not None:
                mapping = self._revision_id_to_revno_cache
            else:
                mapping = self._gen_revno_map()
                self._cache_revision_id_to_revno(mapping)
            # TODO: jam 20070417 Since this is being cached, should we be
            # returning a copy?
            # I would rather not, and instead just declare that users should
            # not modify the return value.
            return mapping

    def _gen_revno_map(self):
        """Create a new mapping from revision ids to dotted revnos.

        Dotted revnos are generated based on the current tip in the revision
        history.
        This is the worker function for get_revision_id_to_revno_map, which
        just caches the return value.

        Returns: A dictionary mapping revision_id => dotted revno.
        """
        revision_id_to_revno = {
            rev_id: revno for rev_id, depth, revno, end_of_merge
            in self.iter_merge_sorted_revisions()}
        return revision_id_to_revno

    def iter_merge_sorted_revisions(self, start_revision_id=None,
                                    stop_revision_id=None,
                                    stop_rule='exclude', direction='reverse'):
        """Walk the revisions for a branch in merge sorted order.

        Merge sorted order is the output from a merge-aware,
        topological sort, i.e. all parents come before their
        children going forward; the opposite for reverse.

        Args:
          start_revision_id: the revision_id to begin walking from.
            If None, the branch tip is used.
          stop_revision_id: the revision_id to terminate the walk
            after. If None, the rest of history is included.
          stop_rule: if stop_revision_id is not None, the precise rule
            to use for termination:

            * 'exclude' - leave the stop revision out of the result (default)
            * 'include' - the stop revision is the last item in the result
            * 'with-merges' - include the stop revision and all of its
              merged revisions in the result
            * 'with-merges-without-common-ancestry' - filter out revisions
              that are in both ancestries
          direction: either 'reverse' or 'forward':

            * reverse means return the start_revision_id first, i.e.
              start at the most recent revision and go backwards in history
            * forward returns tuples in the opposite order to reverse.
              Note in particular that forward does *not* do any intelligent
              ordering w.r.t. depth as some clients of this API may like.
              (If required, that ought to be done at higher layers.)

        Returns: an iterator over (revision_id, depth, revno, end_of_merge)
            tuples where:

            * revision_id: the unique id of the revision
            * depth: How many levels of merging deep this node has been
              found.
            * revno_sequence: This field provides a sequence of
              revision numbers for all revisions. The format is:
              (REVNO, BRANCHNUM, BRANCHREVNO). BRANCHNUM is the number of the
              branch that the revno is on. From left to right the REVNO numbers
              are the sequence numbers within that branch of the revision.
            * end_of_merge: When True the next node (earlier in history) is
              part of a different merge.
        """
        with self.lock_read():
            # Note: depth and revno values are in the context of the branch so
            # we need the full graph to get stable numbers, regardless of the
            # start_revision_id.
            if self._merge_sorted_revisions_cache is None:
                last_revision = self.last_revision()
                known_graph = self.repository.get_known_graph_ancestry(
                    [last_revision])
                self._merge_sorted_revisions_cache = known_graph.merge_sort(
                    last_revision)
            filtered = self._filter_merge_sorted_revisions(
                self._merge_sorted_revisions_cache, start_revision_id,
                stop_revision_id, stop_rule)
            # Make sure we don't return revisions that are not part of the
            # start_revision_id ancestry.
            filtered = self._filter_start_non_ancestors(filtered)
            if direction == 'reverse':
                return filtered
            if direction == 'forward':
                return reversed(list(filtered))
            else:
                raise ValueError('invalid direction %r' % direction)

    def _filter_merge_sorted_revisions(self, merge_sorted_revisions,
                                       start_revision_id, stop_revision_id,
                                       stop_rule):
        """Iterate over an inclusive range of sorted revisions."""
        rev_iter = iter(merge_sorted_revisions)
        if start_revision_id is not None:
            for node in rev_iter:
                rev_id = node.key
                if rev_id != start_revision_id:
                    continue
                else:
                    # The decision to include the start or not
                    # depends on the stop_rule if a stop is provided
                    # so pop this node back into the iterator
                    rev_iter = itertools.chain(iter([node]), rev_iter)
                    break
        if stop_revision_id is None:
            # Yield everything
            for node in rev_iter:
                rev_id = node.key
                yield (rev_id, node.merge_depth, node.revno,
                       node.end_of_merge)
        elif stop_rule == 'exclude':
            for node in rev_iter:
                rev_id = node.key
                if rev_id == stop_revision_id:
                    return
                yield (rev_id, node.merge_depth, node.revno,
                       node.end_of_merge)
        elif stop_rule == 'include':
            for node in rev_iter:
                rev_id = node.key
                yield (rev_id, node.merge_depth, node.revno,
                       node.end_of_merge)
                if rev_id == stop_revision_id:
                    return
        elif stop_rule == 'with-merges-without-common-ancestry':
            # We want to exclude all revisions that are already part of the
            # stop_revision_id ancestry.
            graph = self.repository.get_graph()
            ancestors = graph.find_unique_ancestors(start_revision_id,
                                                    [stop_revision_id])
            for node in rev_iter:
                rev_id = node.key
                if rev_id not in ancestors:
                    continue
                yield (rev_id, node.merge_depth, node.revno,
                       node.end_of_merge)
        elif stop_rule == 'with-merges':
            stop_rev = self.repository.get_revision(stop_revision_id)
            if stop_rev.parent_ids:
                left_parent = stop_rev.parent_ids[0]
            else:
                left_parent = _mod_revision.NULL_REVISION
            # left_parent is the actual revision we want to stop logging at,
            # since we want to show the merged revisions after the stop_rev too
            reached_stop_revision_id = False
            revision_id_whitelist = []
            for node in rev_iter:
                rev_id = node.key
                if rev_id == left_parent:
                    # reached the left parent after the stop_revision
                    return
                if (not reached_stop_revision_id
                        or rev_id in revision_id_whitelist):
                    yield (rev_id, node.merge_depth, node.revno,
                           node.end_of_merge)
                    if reached_stop_revision_id or rev_id == stop_revision_id:
                        # only do the merged revs of rev_id from now on
                        rev = self.repository.get_revision(rev_id)
                        if rev.parent_ids:
                            reached_stop_revision_id = True
                            revision_id_whitelist.extend(rev.parent_ids)
        else:
            raise ValueError('invalid stop_rule %r' % stop_rule)

    def _filter_start_non_ancestors(self, rev_iter):
        # If we started from a dotted revno, we want to consider it as a tip
        # and don't want to yield revisions that are not part of its
        # ancestry. Given the order guaranteed by the merge sort, we will see
        # uninteresting descendants of the first parent of our tip before the
        # tip itself.
        try:
            first = next(rev_iter)
        except StopIteration:
            return
        (rev_id, merge_depth, revno, end_of_merge) = first
        yield first
        if not merge_depth:
            # We start at a mainline revision so by definition, all others
            # revisions in rev_iter are ancestors
            yield from rev_iter

        clean = False
        whitelist = set()
        pmap = self.repository.get_parent_map([rev_id])
        parents = pmap.get(rev_id, [])
        if parents:
            whitelist.update(parents)
        else:
            # If there is no parents, there is nothing of interest left

            # FIXME: It's hard to test this scenario here as this code is never
            # called in that case. -- vila 20100322
            return

        for (rev_id, merge_depth, revno, end_of_merge) in rev_iter:
            if not clean:
                if rev_id in whitelist:
                    pmap = self.repository.get_parent_map([rev_id])
                    parents = pmap.get(rev_id, [])
                    whitelist.remove(rev_id)
                    whitelist.update(parents)
                    if merge_depth == 0:
                        # We've reached the mainline, there is nothing left to
                        # filter
                        clean = True
                else:
                    # A revision that is not part of the ancestry of our
                    # starting revision.
                    continue
            yield (rev_id, merge_depth, revno, end_of_merge)

    def leave_lock_in_place(self):
        """Tell this branch object not to release the physical lock when this
        object is unlocked.

        If lock_write doesn't return a token, then this method is not
        supported.
        """
        self.control_files.leave_in_place()

    def dont_leave_lock_in_place(self):
        """Tell this branch object to release the physical lock when this
        object is unlocked, even if it didn't originally acquire it.

        If lock_write doesn't return a token, then this method is not
        supported.
        """
        self.control_files.dont_leave_in_place()

    def bind(self, other):
        """Bind the local branch the other branch.

        Args:
          other: The branch to bind to
        """
        raise BindingUnsupported(self)

    def get_append_revisions_only(self):
        """Whether it is only possible to append revisions to the history.
        """
        if not self._format.supports_set_append_revisions_only():
            return False
        return self.get_config_stack().get('append_revisions_only')

    def set_append_revisions_only(self, enabled: bool) -> None:
        if not self._format.supports_set_append_revisions_only():
            raise errors.UpgradeRequired(self.user_url)
        self.get_config_stack().set('append_revisions_only', enabled)

    def fetch(self, from_branch, stop_revision=None, limit=None, lossy=False):
        """Copy revisions from from_branch into this branch.

        Args:
          from_branch: Where to copy from.
          stop_revision: What revision to stop at (None for at the end
                              of the branch.
          limit: Optional rough limit of revisions to fetch
        Returns: None
        """
        with self.lock_write():
            return InterBranch.get(from_branch, self).fetch(
                stop_revision, limit=limit, lossy=lossy)

    def get_bound_location(self) -> Optional[str]:
        """Return the URL of the branch we are bound to.

        Older format branches cannot bind, please be sure to use a metadir
        branch.
        """
        return None

    def get_old_bound_location(self):
        """Return the URL of the branch we used to be bound to
        """
        raise errors.UpgradeRequired(self.user_url)

    def get_commit_builder(self, parents, config_stack=None, timestamp=None,
                           timezone=None, committer=None, revprops=None,
                           revision_id=None, lossy=False):
        """Obtain a CommitBuilder for this branch.

        Args:
          parents: Revision ids of the parents of the new revision.
          config: Optional configuration to use.
          timestamp: Optional timestamp recorded for commit.
          timezone: Optional timezone for timestamp.
          committer: Optional committer to set for commit.
          revprops: Optional dictionary of revision properties.
          revision_id: Optional revision id.
          lossy: Whether to discard data that can not be natively
            represented, when pushing to a foreign VCS
        """

        if config_stack is None:
            config_stack = self.get_config_stack()

        return self.repository.get_commit_builder(
            self, parents, config_stack, timestamp, timezone, committer,
            revprops, revision_id, lossy)

    def get_master_branch(
            self, possible_transports: Optional[List[Transport]] = None
            ) -> Optional["Branch"]:
        """Return the branch we are bound to.

        Returns: Either a Branch, or None
        """
        return None

    def get_stacked_on_url(self) -> str:
        """Get the URL this branch is stacked against.

        Raises:
          NotStacked: If the branch is not stacked.
          UnstackableBranchFormat: If the branch does not support
            stacking.
        """
        raise NotImplementedError(self.get_stacked_on_url)

    def set_last_revision_info(
            self, revno: Optional[int], revision_id: RevisionID) -> None:
        """Set the last revision of this branch.

        The caller is responsible for checking that the revno is correct
        for this revision id.

        It may be possible to set the branch last revision to an id not
        present in the repository.  However, branches can also be
        configured to check constraints on history, in which case this may not
        be permitted.
        """
        raise NotImplementedError(self.set_last_revision_info)

    def generate_revision_history(self, revision_id: RevisionID,
                                  last_rev: Optional[RevisionID] = None,
                                  other_branch: Optional["Branch"] = None) -> None:
        """See Branch.generate_revision_history"""
        with self.lock_write():
            graph = self.repository.get_graph()
            (last_revno, last_revid) = self.last_revision_info()
            known_revision_ids = [
                (last_revid, last_revno),
                (_mod_revision.NULL_REVISION, 0),
                ]
            if last_rev is not None:
                if not graph.is_ancestor(last_rev, revision_id):
                    # our previous tip is not merged into stop_revision
                    raise errors.DivergedBranches(self, other_branch)
            revno = graph.find_distance_to_null(
                revision_id, known_revision_ids)
            self.set_last_revision_info(revno, revision_id)

    def _set_parent_location(self, url: Optional[str]) -> None:
        raise NotImplementedError(self._set_parent_location)

    def set_parent(self, url: Optional[str]) -> None:
        """See Branch.set_parent."""
        # TODO: Maybe delete old location files?
        # URLs should never be unicode, even on the local fs,
        # FIXUP this and get_parent in a future branch format bump:
        # read and rewrite the file. RBC 20060125
        if url is not None:
            if isinstance(url, str):
                try:
                    url.encode('ascii')
                except UnicodeEncodeError as exc:
                    raise urlutils.InvalidURL(
                        url, "Urls must be 7-bit ascii, "
                        "use breezy.urlutils.escape") from exc
            url = urlutils.relative_url(self.base, url)
        with self.lock_write():
            self._set_parent_location(url)

    def set_stacked_on_url(self, url: str) -> None:
        """Set the URL this branch is stacked against.

        :raises UnstackableBranchFormat: If the branch does not support
            stacking.
        :raises UnstackableRepositoryFormat: If the repository does not support
            stacking.
        """
        raise NotImplementedError(self.set_stacked_on_url)

    def _cache_revision_history(self, rev_history):
        """Set the cached revision history to rev_history.

        The revision_history method will use this cache to avoid regenerating
        the revision history.

        This API is semi-public; it only for use by subclasses, all other code
        should consider it to be private.
        """
        self._revision_history_cache = rev_history

    def _cache_revision_id_to_revno(self, revision_id_to_revno):
        """Set the cached revision_id => revno map to revision_id_to_revno.

        This API is semi-public; it only for use by subclasses, all other code
        should consider it to be private.
        """
        self._revision_id_to_revno_cache = revision_id_to_revno

    def _clear_cached_state(self) -> None:
        """Clear any cached data on this branch, e.g. cached revision history.

        This means the next call to revision_history will need to call
        _gen_revision_history.

        This API is semi-public; it is only for use by subclasses, all other
        code should consider it to be private.
        """
        self._revision_history_cache = None
        self._revision_id_to_revno_cache = None
        self._last_revision_info_cache = None
        self._master_branch_cache = None
        self._merge_sorted_revisions_cache = None
        self._partial_revision_history_cache = []
        self._partial_revision_id_to_revno_cache = {}

    def _gen_revision_history(self):
        """Return sequence of revision hashes on to this branch.

        Unlike revision_history, this method always regenerates or rereads the
        revision history, i.e. it does not cache the result, so repeated calls
        may be expensive.

        Concrete subclasses should override this instead of revision_history so
        that subclasses do not need to deal with caching logic.

        This API is semi-public; it only for use by subclasses, all other code
        should consider it to be private.
        """
        raise NotImplementedError(self._gen_revision_history)

    def _revision_history(self) -> List[RevisionID]:
        if 'evil' in debug.debug_flags:
            mutter_callsite(3, "revision_history scales with history.")
        if self._revision_history_cache is not None:
            history = self._revision_history_cache
        else:
            history = self._gen_revision_history()
            self._cache_revision_history(history)
        return list(history)

    def revno(self):
        """Return current revision number for this branch.

        That is equivalent to the number of revisions committed to
        this branch.
        """
        return self.last_revision_info()[0]

    def unbind(self):
        """Older format branches cannot bind or unbind."""
        raise errors.UpgradeRequired(self.user_url)

    def last_revision(self) -> RevisionID:
        """Return last revision id, or NULL_REVISION."""
        return self.last_revision_info()[1]

    def last_revision_info(self) -> Tuple[int, RevisionID]:
        """Return information about the last revision.

        Returns: A tuple (revno, revision_id).
        """
        with self.lock_read():
            if self._last_revision_info_cache is None:
                self._last_revision_info_cache = (
                    self._read_last_revision_info())
            return self._last_revision_info_cache

    def _read_last_revision_info(self):
        raise NotImplementedError(self._read_last_revision_info)

    def import_last_revision_info_and_tags(self, source, revno, revid,
                                           *, lossy=False):
        """Set the last revision info, importing from another repo if necessary.

        This is used by the bound branch code to upload a revision to
        the master branch first before updating the tip of the local branch.
        Revisions referenced by source's tags are also transferred.

        Args:
          source: Source branch to optionally fetch from
          revno: Revision number of the new tip
          revid: Revision id of the new tip
          lossy: Whether to discard metadata that can not be
            natively represented
        Returns: Tuple with the new revision number and revision id
            (should only be different from the arguments when lossy=True)
        """
        if not self.repository.has_same_location(source.repository):
            self.fetch(source, revid)
        self.set_last_revision_info(revno, revid)
        return (revno, revid)

    def revision_id_to_revno(self, revision_id: RevisionID) -> int:
        """Given a revision id, return its revno"""
        if _mod_revision.is_null(revision_id):
            return 0
        history = self._revision_history()
        try:
            return history.index(revision_id) + 1
        except ValueError as exc:
            raise errors.NoSuchRevision(self, revision_id) from exc

    def get_rev_id(self, revno: int, history: Optional[List[RevisionID]] = None) -> RevisionID:
        """Find the revision id of the specified revno."""
        with self.lock_read():
            if revno == 0:
                return _mod_revision.NULL_REVISION
            last_revno, last_revid = self.last_revision_info()
            if revno == last_revno:
                return last_revid
            if revno <= 0 or revno > last_revno:
                raise errors.NoSuchRevision(self, revno)
            distance_from_last = last_revno - revno
            if len(self._partial_revision_history_cache) <= distance_from_last:
                self._extend_partial_history(distance_from_last)
            return self._partial_revision_history_cache[distance_from_last]

    def pull(self, source: "Branch", *, overwrite: bool = False,
             stop_revision: Optional[RevisionID] = None,
             possible_transports: Optional[List[Transport]] = None,
             **kwargs) -> "PullResult":
        """Mirror source into this branch.

        This branch is considered to be 'local', having low latency.

        Returns: PullResult instance
        """
        return InterBranch.get(source, self).pull(
            overwrite=overwrite, stop_revision=stop_revision,
            possible_transports=possible_transports, **kwargs)

    def push(self, target: "Branch", *, overwrite: bool = False,
             stop_revision: Optional[RevisionID] = None, lossy: bool = False,
             **kwargs):
        """Mirror this branch into target.

        This branch is considered to be 'local', having low latency.
        """
        return InterBranch.get(self, target).push(
            overwrite, stop_revision, lossy, **kwargs)

    def basis_tree(self):
        """Return `Tree` object for last revision."""
        return self.repository.revision_tree(self.last_revision())

    def get_parent(self) -> Optional[str]:
        """Return the parent location of the branch.

        This is the default location for pull/missing.  The usual
        pattern is that the user can override it by specifying a
        location.
        """
        parent = self._get_parent_location()
        if parent is None:
            return parent
        # This is an old-format absolute path to a local branch
        # turn it into a url
        if parent.startswith('/'):
            parent = urlutils.local_path_to_url(parent)
        try:
            return urlutils.join(self.base[:-1], parent)
        except urlutils.InvalidURLJoin as exc:
            raise errors.InaccessibleParent(parent, self.user_url) from exc

    def _get_parent_location(self):
        raise NotImplementedError(self._get_parent_location)

    def _set_config_location(self, name, url, *, config=None,
                             make_relative=False):
        if config is None:
            config = self.get_config_stack()
        if url is None:
            url = ''
        elif make_relative:
            url = urlutils.relative_url(self.base, url)
        config.set(name, url)

    def _get_config_location(self, name: str, *, config=None) -> Optional[str]:
        if config is None:
            config = self.get_config_stack()
        location = config.get(name)
        if location == '':
            location = None
        return cast(Optional[str], location)

    def get_child_submit_format(self) -> Optional[str]:
        """Return the preferred format of submissions to this branch."""
        return cast(Optional[str], self.get_config_stack().get('child_submit_format'))

    def get_submit_branch(self) -> Optional[str]:
        """Return the submit location of the branch.

        This is the default location for bundle.  The usual
        pattern is that the user can override it by specifying a
        location.
        """
        return cast(Optional[str], self.get_config_stack().get('submit_branch'))

    def set_submit_branch(self, location: str) -> None:
        """Return the submit location of the branch.

        This is the default location for bundle.  The usual
        pattern is that the user can override it by specifying a
        location.
        """
        self.get_config_stack().set('submit_branch', location)

    def get_public_branch(self) -> Optional[str]:
        """Return the public location of the branch.

        This is used by merge directives.
        """
        return self._get_config_location('public_branch')

    def set_public_branch(self, location: str) -> None:
        """Return the submit location of the branch.

        This is the default location for bundle.  The usual
        pattern is that the user can override it by specifying a
        location.
        """
        self._set_config_location('public_branch', location)

    def get_push_location(self) -> Optional[str]:
        """Return None or the location to push this branch to."""
        return cast(str, self.get_config_stack().get('push_location'))

    def set_push_location(self, location: str) -> None:
        """Set a new push location for this branch."""
        raise NotImplementedError(self.set_push_location)

    def _run_post_change_branch_tip_hooks(self, old_revno, old_revid):
        """Run the post_change_branch_tip hooks."""
        hooks = Branch.hooks['post_change_branch_tip']
        if not hooks:
            return
        new_revno, new_revid = self.last_revision_info()
        params = ChangeBranchTipParams(
            self, old_revno, new_revno, old_revid, new_revid)
        for hook in hooks:
            hook(params)

    def _run_pre_change_branch_tip_hooks(self, new_revno, new_revid):
        """Run the pre_change_branch_tip hooks."""
        hooks = Branch.hooks['pre_change_branch_tip']
        if not hooks:
            return
        old_revno, old_revid = self.last_revision_info()
        params = ChangeBranchTipParams(
            self, old_revno, new_revno, old_revid, new_revid)
        for hook in hooks:
            hook(params)

    def update(self) -> None:
        """Synchronise this branch with the master branch if any.

        Returns: None or the last_revision pivoted out during the update.
        """
        return None

    def check_revno(self, revno: int) -> None:
        """\
        Check whether a revno corresponds to any revision.
        Zero (the NULL revision) is considered valid.
        """
        if revno != 0:
            self.check_real_revno(revno)

    def check_real_revno(self, revno: int) -> None:
        """\
        Check whether a revno corresponds to a real revision.
        Zero (the NULL revision) is considered invalid
        """
        if revno < 1 or revno > self.revno():
            raise errors.InvalidRevisionNumber(revno)

    def clone(self, to_controldir: ControlDir, *,
              revision_id: Optional[RevisionID] = None, name: Optional[str] = None,
              repository_policy=None, tag_selector=None) -> "Branch":
        """Clone this branch into to_controldir preserving all semantic values.

        Most API users will want 'create_clone_on_transport', which creates a
        new bzrdir and branch on the fly.

        revision_id: if not None, the revision history in the new branch will
                     be truncated to end with revision_id.
        """
        result = to_controldir.create_branch(name=name)
        with self.lock_read(), result.lock_write():
            if repository_policy is not None:
                repository_policy.configure_branch(result)
            self.copy_content_into(
                result, revision_id=revision_id, tag_selector=tag_selector)
        return result

    def sprout(self, to_controldir, *, revision_id=None, repository_policy=None,
               repository=None, lossy=False, tag_selector=None,
               name=None):
        """Create a new line of development from the branch, into to_controldir.

        to_controldir controls the branch format.

        revision_id: if not None, the revision history in the new branch will
                     be truncated to end with revision_id.
        """
        if (repository_policy is not None
                and repository_policy.requires_stacking()):
            to_controldir._format.require_stacking(_skip_repo=True)
        result = to_controldir.create_branch(repository=repository, name=name)
        if lossy:
            raise errors.LossyPushToSameVCS(self, result)
        with self.lock_read(), result.lock_write():
            if repository_policy is not None:
                repository_policy.configure_branch(result)
            self.copy_content_into(
                result, revision_id=revision_id, tag_selector=tag_selector)
            master_url = self.get_bound_location()
            if master_url is None:
                result.set_parent(self.user_url)
            else:
                result.set_parent(master_url)
        return result

    def _synchronize_history(self, destination, revision_id):
        """Synchronize last revision and revision history between branches.

        This version is most efficient when the destination is also a
        BzrBranch6, but works for BzrBranch5, as long as the destination's
        repository contains all the lefthand ancestors of the intended
        last_revision.  If not, set_last_revision_info will fail.

        Args:
          destination: The branch to copy the history into
          revision_id: The revision-id to truncate history at.  May
             be None to copy complete history.
        """
        source_revno, source_revision_id = self.last_revision_info()
        if revision_id is None:
            revno, revision_id = source_revno, source_revision_id
        else:
            graph = self.repository.get_graph()
            try:
                revno = graph.find_distance_to_null(
                    revision_id, [(source_revision_id, source_revno)])
            except errors.GhostRevisionsHaveNoRevno:
                # Default to 1, if we can't find anything else
                revno = 1
        destination.set_last_revision_info(revno, revision_id)

    def copy_content_into(self, destination, *, revision_id=None, tag_selector=None):
        """Copy the content of self into destination.

        revision_id: if not None, the revision history in the new branch will
                     be truncated to end with revision_id.
        tag_selector: Optional callback that receives a tag name
            and should return a boolean to indicate whether a tag should be copied
        """
        return InterBranch.get(self, destination).copy_content_into(
            revision_id=revision_id, tag_selector=tag_selector)

    def update_references(self, target):
        if not self._format.supports_reference_locations:
            return
        return InterBranch.get(self, target).update_references()

    def check(self, refs):
        """Check consistency of the branch.

        In particular this checks that revisions given in the revision-history
        do actually match up in the revision graph, and that they're all
        present in the repository.

        Callers will typically also want to check the repository.

        Args:
          refs: Calculated refs for this branch as specified by
            branch._get_check_refs()
        Returns: A BranchCheckResult.
        """
        with self.lock_read():
            result = BranchCheckResult(self)
            last_revno, last_revision_id = self.last_revision_info()
            actual_revno = refs[('lefthand-distance', last_revision_id)]
            if actual_revno != last_revno:
                result.errors.append(errors.BzrCheckError(
                    'revno does not match len(mainline) {} != {}'.format(
                        last_revno, actual_revno)))
            # TODO: We should probably also check that self.revision_history
            # matches the repository for older branch formats.
            # If looking for the code that cross-checks repository parents
            # against the Graph.iter_lefthand_ancestry output, that is now a
            # repository specific check.
            return result

    def _get_checkout_format(self, lightweight=False):
        """Return the most suitable metadir for a checkout of this branch.
        Weaves are used if this branch's repository uses weaves.
        """
        format = self.repository.controldir.checkout_metadir()
        format.set_branch_format(self._format)
        return format

    def create_clone_on_transport(self, to_transport, *, revision_id=None,
                                  stacked_on=None, create_prefix=False,
                                  use_existing_dir=False, no_tree=None,
                                  tag_selector=None):
        """Create a clone of this branch and its bzrdir.

        Args:
          to_transport: The transport to clone onto.
          revision_id: The revision id to use as tip in the new branch.
            If None the tip is obtained from this branch.
          stacked_on: An optional URL to stack the clone on.
          create_prefix: Create any missing directories leading up to
            to_transport.
          use_existing_dir: Use an existing directory if one exists.
        """
        # XXX: Fix the bzrdir API to allow getting the branch back from the
        # clone call. Or something. 20090224 RBC/spiv.
        # XXX: Should this perhaps clone colocated branches as well,
        # rather than just the default branch? 20100319 JRV
        if revision_id is None:
            revision_id = self.last_revision()
        dir_to = self.controldir.clone_on_transport(
            to_transport, revision_id=revision_id, stacked_on=stacked_on,
            create_prefix=create_prefix, use_existing_dir=use_existing_dir,
            no_tree=no_tree, tag_selector=tag_selector)
        return dir_to.open_branch()

    def create_checkout(self, to_location, *, revision_id=None,
                        lightweight=False, accelerator_tree=None,
                        hardlink=False, recurse_nested=True):
        """Create a checkout of a branch.

        Args:
          to_location: The url to produce the checkout at
          revision_id: The revision to check out
          lightweight: If True, produce a lightweight checkout, otherwise,
            produce a bound branch (heavyweight checkout)
          accelerator_tree: A tree which can be used for retrieving file
            contents more quickly than the revision tree, i.e. a workingtree.
            The revision tree will be used for cases where accelerator_tree's
            content is different.
          hardlink: If true, hard-link files from accelerator_tree,
            where possible.
          recurse_nested: Whether to recurse into nested trees
        Returns: The tree of the created checkout
        """
        t = get_transport(to_location)
        t.ensure_base()
        format = self._get_checkout_format(lightweight=lightweight)
        try:
            checkout = format.initialize_on_transport(t)
        except errors.AlreadyControlDirError as exc:
            # It's fine if the control directory already exists,
            # as long as there is no existing branch and working tree.
            checkout = ControlDir.open_from_transport(t)
            try:
                checkout.open_branch()
            except errors.NotBranchError:
                pass
            else:
                raise errors.AlreadyControlDirError(t.base) from exc
            if (checkout.control_transport.base
                    == self.controldir.control_transport.base):
                # When checking out to the same control directory,
                # always create a lightweight checkout
                lightweight = True

        if lightweight:
            from_branch = checkout.set_branch_reference(target_branch=self)
        else:
            policy = checkout.determine_repository_policy()
            policy.acquire_repository()
            checkout_branch = checkout.create_branch()
            checkout_branch.bind(self)
            # pull up to the specified revision_id to set the initial
            # branch tip correctly, and seed it with history.
            checkout_branch.pull(self, stop_revision=revision_id)
            from_branch = None
        tree = checkout.create_workingtree(revision_id,
                                           from_branch=from_branch,
                                           accelerator_tree=accelerator_tree,
                                           hardlink=hardlink)
        basis_tree = tree.basis_tree()
        with basis_tree.lock_read():
            for path in basis_tree.iter_references():
                reference_parent = tree.reference_parent(path)
                if reference_parent is None:
                    warning('Branch location for %s unknown.', path)
                    continue
                reference_parent.create_checkout(
                    tree.abspath(path),
                    revision_id=basis_tree.get_reference_revision(path),
                    lightweight=lightweight)
        return tree

    def reconcile(self, thorough=True):
        """Make sure the data stored in this branch is consistent.

        Returns: A `ReconcileResult` object.
        """
        raise NotImplementedError(self.reconcile)

    def supports_tags(self):
        return self._format.supports_tags()

    def automatic_tag_name(self, revision_id):
        """Try to automatically find the tag name for a revision.

        Args:
          revision_id: Revision id of the revision.
        Returns: A tag name or None if no tag name could be determined.
        """
        for hook in Branch.hooks['automatic_tag_name']:
            ret = hook(self, revision_id)
            if ret is not None:
                return ret
        return None

    def _check_if_descendant_or_diverged(self, revision_a, revision_b, graph,
                                         other_branch):
        """Ensure that revision_b is a descendant of revision_a.

        This is a helper function for update_revisions.

        :raises: DivergedBranches if revision_b has diverged from revision_a.
        Returns: True if revision_b is a descendant of revision_a.
        """
        relation = self._revision_relations(revision_a, revision_b, graph)
        if relation == 'b_descends_from_a':
            return True
        elif relation == 'diverged':
            raise errors.DivergedBranches(self, other_branch)
        elif relation == 'a_descends_from_b':
            return False
        else:
            raise AssertionError("invalid relation: {!r}".format(relation))

    def _revision_relations(self, revision_a, revision_b, graph):
        """Determine the relationship between two revisions.

        Returns: One of: 'a_descends_from_b', 'b_descends_from_a', 'diverged'
        """
        heads = graph.heads([revision_a, revision_b])
        if heads == {revision_b}:
            return 'b_descends_from_a'
        elif heads == {revision_a, revision_b}:
            # These branches have diverged
            return 'diverged'
        elif heads == {revision_a}:
            return 'a_descends_from_b'
        else:
            raise AssertionError("invalid heads: {!r}".format(heads))

    def heads_to_fetch(self):
        """Return the heads that must and that should be fetched to copy this
        branch into another repo.

        Returns: a 2-tuple of (must_fetch, if_present_fetch).  must_fetch is a
            set of heads that must be fetched.  if_present_fetch is a set of
            heads that must be fetched if present, but no error is necessary if
            they are not present.
        """
        # For bzr native formats must_fetch is just the tip, and
        # if_present_fetch are the tags.
        must_fetch = {self.last_revision()}
        if_present_fetch = set()
        if self.get_config_stack().get('branch.fetch_tags'):
            try:
                if_present_fetch = set(self.tags.get_reverse_tag_dict())
            except errors.TagsNotSupported:
                pass
        must_fetch.discard(_mod_revision.NULL_REVISION)
        if_present_fetch.discard(_mod_revision.NULL_REVISION)
        return must_fetch, if_present_fetch

    def create_memorytree(self):
        """Create a memory tree for this branch.

        Returns: An in-memory MutableTree instance
        """
        from . import memorytree
        return memorytree.MemoryTree.create_on_branch(self)


class BranchFormat(ControlComponentFormat):
    """An encapsulation of the initialization and open routines for a format.

    Formats provide three things:
     * An initialization routine,
     * a format description
     * an open routine.

    Formats are placed in an dict by their format string for reference
    during branch opening. It's not required that these be instances, they
    can be classes themselves with class methods - it simply depends on
    whether state is needed for a given format or not.

    Once a format is deprecated, just deprecate the initialize and open
    methods on the format class. Do not deprecate the object, as the
    object will be created every time regardless.
    """

    def __eq__(self, other):
        return self.__class__ is other.__class__

    def __ne__(self, other):
        return not (self == other)

    def get_reference(self, controldir, name=None):
        """Get the target reference of the branch in controldir.

        format probing must have been completed before calling
        this method - it is assumed that the format of the branch
        in controldir is correct.

        Args:
          controldir: The controldir to get the branch data from.
          name: Name of the colocated branch to fetch
        Returns: None if the branch is not a reference branch.
        """
        return None

    @classmethod
    def set_reference(self, controldir, name, to_branch):
        """Set the target reference of the branch in controldir.

        format probing must have been completed before calling
        this method - it is assumed that the format of the branch
        in controldir is correct.

        Args:
          controldir: The controldir to set the branch reference for.
          name: Name of colocated branch to set, None for default
          to_branch: branch that the checkout is to reference
        """
        raise NotImplementedError(self.set_reference)

    def get_format_description(self):
        """Return the short format description for this format."""
        raise NotImplementedError(self.get_format_description)

    def _run_post_branch_init_hooks(self, controldir, name, branch):
        hooks = Branch.hooks['post_branch_init']
        if not hooks:
            return
        params = BranchInitHookParams(self, controldir, name, branch)
        for hook in hooks:
            hook(params)

    def initialize(self, controldir, name=None, repository=None,
                   append_revisions_only=None):
        """Create a branch of this format in controldir.

        Args:
          name: Name of the colocated branch to create.
        """
        raise NotImplementedError(self.initialize)

    def is_supported(self):
        """Is this format supported?

        Supported formats can be initialized and opened.
        Unsupported formats may not support initialization or committing or
        some other features depending on the reason for not being supported.
        """
        return True

    def make_tags(self, branch):
        """Create a tags object for branch.

        This method is on BranchFormat, because BranchFormats are reflected
        over the wire via network_name(), whereas full Branch instances require
        multiple VFS method calls to operate at all.

        The default implementation returns a disabled-tags instance.

        Note that it is normal for branch to be a RemoteBranch when using tags
        on a RemoteBranch.
        """
        from .tag import DisabledTags
        return DisabledTags(branch)

    def network_name(self):
        """A simple byte string uniquely identifying this format for RPC calls.

        MetaDir branch formats use their disk format string to identify the
        repository over the wire. All in one formats such as bzr < 0.8, and
        foreign formats like svn/git and hg should use some marker which is
        unique and immutable.
        """
        raise NotImplementedError(self.network_name)

    def open(self, controldir, name=None, _found=False, ignore_fallbacks=False,
             found_repository=None, possible_transports=None):
        """Return the branch object for controldir.

        Args:
          controldir: A ControlDir that contains a branch.
          name: Name of colocated branch to open
          _found: a private parameter, do not use it. It is used to
            indicate if format probing has already be done.
          ignore_fallbacks: when set, no fallback branches will be opened
            (if there are any).  Default is to open fallbacks.
        """
        raise NotImplementedError(self.open)

    def supports_set_append_revisions_only(self):
        """True if this format supports set_append_revisions_only."""
        return False

    def supports_stacking(self):
        """True if this format records a stacked-on branch."""
        return False

    def supports_leaving_lock(self):
        """True if this format supports leaving locks in place."""
        return False  # by default

    def __str__(self):
        return self.get_format_description().rstrip()

    def supports_tags(self):
        """True if this format supports tags stored in the branch"""
        return False  # by default

    def tags_are_versioned(self):
        """Whether the tag container for this branch versions tags."""
        return False

    def supports_tags_referencing_ghosts(self):
        """True if tags can reference ghost revisions."""
        return True

    def supports_store_uncommitted(self):
        """True if uncommitted changes can be stored in this branch."""
        return True

    def stores_revno(self):
        """True if this branch format store revision numbers."""
        return True


class BranchHooks(Hooks):
    """A dictionary mapping hook name to a list of callables for branch hooks.

    e.g. ['post_push'] Is the list of items to be called when the
    push function is invoked.
    """

    def __init__(self):
        """Create the default hooks.

        These are all empty initially, because by default nothing should get
        notified.
        """
        Hooks.__init__(self, "breezy.branch", "Branch.hooks")
        self.add_hook(
            'open',
            "Called with the Branch object that has been opened after a "
            "branch is opened.", (1, 8))
        self.add_hook(
            'post_push',
            "Called after a push operation completes. post_push is called "
            "with a breezy.branch.BranchPushResult object and only runs in "
            "the bzr client.", (0, 15))
        self.add_hook(
            'post_pull',
            "Called after a pull operation completes. post_pull is called "
            "with a breezy.branch.PullResult object and only runs in the "
            "bzr client.", (0, 15))
        self.add_hook(
            'pre_commit',
            "Called after a commit is calculated but before it is "
            "completed. pre_commit is called with (local, master, old_revno, "
            "old_revid, future_revno, future_revid, tree_delta, future_tree"
            "). old_revid is NULL_REVISION for the first commit to a branch, "
            "tree_delta is a TreeDelta object describing changes from the "
            "basis revision. hooks MUST NOT modify this delta. "
            " future_tree is an in-memory tree obtained from "
            "CommitBuilder.revision_tree() and hooks MUST NOT modify this "
            "tree.", (0, 91))
        self.add_hook(
            'post_commit',
            "Called in the bzr client after a commit has completed. "
            "post_commit is called with (local, master, old_revno, old_revid, "
            "new_revno, new_revid). old_revid is NULL_REVISION for the first "
            "commit to a branch.", (0, 15))
        self.add_hook(
            'post_uncommit',
            "Called in the bzr client after an uncommit completes. "
            "post_uncommit is called with (local, master, old_revno, "
            "old_revid, new_revno, new_revid) where local is the local branch "
            "or None, master is the target branch, and an empty branch "
            "receives new_revno of 0, new_revid of None.", (0, 15))
        self.add_hook(
            'pre_change_branch_tip',
            "Called in bzr client and server before a change to the tip of a "
            "branch is made. pre_change_branch_tip is called with a "
            "breezy.branch.ChangeBranchTipParams. Note that push, pull, "
            "commit, uncommit will all trigger this hook.", (1, 6))
        self.add_hook(
            'post_change_branch_tip',
            "Called in bzr client and server after a change to the tip of a "
            "branch is made. post_change_branch_tip is called with a "
            "breezy.branch.ChangeBranchTipParams. Note that push, pull, "
            "commit, uncommit will all trigger this hook.", (1, 4))
        self.add_hook(
            'transform_fallback_location',
            "Called when a stacked branch is activating its fallback "
            "locations. transform_fallback_location is called with (branch, "
            "url), and should return a new url. Returning the same url "
            "allows it to be used as-is, returning a different one can be "
            "used to cause the branch to stack on a closer copy of that "
            "fallback_location. Note that the branch cannot have history "
            "accessing methods called on it during this hook because the "
            "fallback locations have not been activated. When there are "
            "multiple hooks installed for transform_fallback_location, "
            "all are called with the url returned from the previous hook."
            "The order is however undefined.", (1, 9))
        self.add_hook(
            'automatic_tag_name',
            "Called to determine an automatic tag name for a revision. "
            "automatic_tag_name is called with (branch, revision_id) and "
            "should return a tag name or None if no tag name could be "
            "determined. The first non-None tag name returned will be used.",
            (2, 2))
        self.add_hook(
            'post_branch_init',
            "Called after new branch initialization completes. "
            "post_branch_init is called with a "
            "breezy.branch.BranchInitHookParams. "
            "Note that init, branch and checkout (both heavyweight and "
            "lightweight) will all trigger this hook.", (2, 2))
        self.add_hook(
            'post_switch',
            "Called after a checkout switches branch. "
            "post_switch is called with a "
            "breezy.branch.SwitchHookParams.", (2, 2))


# install the default hooks into the Branch class.
Branch.hooks = BranchHooks()  # type: ignore


class ChangeBranchTipParams:
    """Object holding parameters passed to `*_change_branch_tip` hooks.

    There are 5 fields that hooks may wish to access:

    Attributes:
      branch: the branch being changed
      old_revno: revision number before the change
      new_revno: revision number after the change
      old_revid: revision id before the change
      new_revid: revision id after the change

    The revid fields are strings. The revno fields are integers.
    """

    def __init__(self, branch, old_revno, new_revno, old_revid, new_revid):
        """Create a group of ChangeBranchTip parameters.

        Args:
          branch: The branch being changed.
          old_revno: Revision number before the change.
          new_revno: Revision number after the change.
          old_revid: Tip revision id before the change.
          new_revid: Tip revision id after the change.
        """
        self.branch = branch
        self.old_revno = old_revno
        self.new_revno = new_revno
        self.old_revid = old_revid
        self.new_revid = new_revid

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __repr__(self):
        return "<{} of {} from ({}, {}) to ({}, {})>".format(
            self.__class__.__name__, self.branch,
            self.old_revno, self.old_revid, self.new_revno, self.new_revid)


class BranchInitHookParams:
    """Object holding parameters passed to `*_branch_init` hooks.

    There are 4 fields that hooks may wish to access:

    Attributes:
      format: the branch format
      bzrdir: the ControlDir where the branch will be/has been initialized
      name: name of colocated branch, if any (or None)
      branch: the branch created

    Note that for lightweight checkouts, the bzrdir and format fields refer to
    the checkout, hence they are different from the corresponding fields in
    branch, which refer to the original branch.
    """

    def __init__(self, format, controldir, name, branch):
        """Create a group of BranchInitHook parameters.

        Args:
          format: the branch format
          controldir: the ControlDir where the branch will be/has been
            initialized
          name: name of colocated branch, if any (or None)
          branch: the branch created

        Note that for lightweight checkouts, the bzrdir and format fields refer
        to the checkout, hence they are different from the corresponding fields
        in branch, which refer to the original branch.
        """
        self.format = format
        self.controldir = controldir
        self.name = name
        self.branch = branch

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __repr__(self):
        return "<{} of {}>".format(self.__class__.__name__, self.branch)


class SwitchHookParams:
    """Object holding parameters passed to `*_switch` hooks.

    There are 4 fields that hooks may wish to access:

    Attributes:
      control_dir: ControlDir of the checkout to change
      to_branch: branch that the checkout is to reference
      force: skip the check for local commits in a heavy checkout
      revision_id: revision ID to switch to (or None)
    """

    def __init__(self, control_dir, to_branch, force, revision_id):
        """Create a group of SwitchHook parameters.

        Args:
          control_dir: ControlDir of the checkout to change
          to_branch: branch that the checkout is to reference
          force: skip the check for local commits in a heavy checkout
          revision_id: revision ID to switch to (or None)
        """
        self.control_dir = control_dir
        self.to_branch = to_branch
        self.force = force
        self.revision_id = revision_id

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __repr__(self):
        return "<{} for {} to ({}, {})>".format(
            self.__class__.__name__, self.control_dir, self.to_branch,
            self.revision_id)


class BranchFormatRegistry(ControlComponentFormatRegistry):
    """Branch format registry."""

    def __init__(self, other_registry=None):
        super().__init__(other_registry)
        self._default_format = None
        self._default_format_key = None

    def get_default(self):
        """Return the current default format."""
        if (self._default_format_key is not None
                and self._default_format is None):
            self._default_format = self.get(self._default_format_key)
        return self._default_format

    def set_default(self, format):
        """Set the default format."""
        self._default_format = format
        self._default_format_key = None

    def set_default_key(self, format_string):
        """Set the default format by its format string."""
        self._default_format_key = format_string
        self._default_format = None


network_format_registry = registry.FormatRegistry[BranchFormat]()
"""Registry of formats indexed by their network name.

The network name for a branch format is an identifier that can be used when
referring to formats with smart server operations. See
BranchFormat.network_name() for more detail.
"""

format_registry = BranchFormatRegistry(network_format_registry)


class BranchWriteLockResult(LogicalLockResult):
    """The result of write locking a branch.

    Attributes:
      token: The token obtained from the underlying branch lock, or
        None.
      unlock: A callable which will unlock the lock.
    """

    def __repr__(self):
        return "BranchWriteLockResult({!r}, {!r})".format(self.unlock, self.token)


######################################################################
# results of operations


class _Result:

    def _show_tag_conficts(self, to_file):
        if not getattr(self, 'tag_conflicts', None):
            return
        to_file.write('Conflicting tags:\n')
        for name, value1, value2 in self.tag_conflicts:
            to_file.write('    {}\n'.format(name))


class PullResult(_Result):
    """Result of a Branch.pull operation.

    Attributes:
      old_revno: Revision number before pull.
      new_revno: Revision number after pull.
      old_revid: Tip revision id before pull.
      new_revid: Tip revision id after pull.
      source_branch: Source (local) branch object. (read locked)
      master_branch: Master branch of the target, or the target if no
        Master
      local_branch: target branch if there is a Master, else None
      target_branch: Target/destination branch object. (write locked)
      tag_conflicts: A list of tag conflicts, see BasicTags.merge_to
      tag_updates: A dict with new tags, see BasicTags.merge_to
    """

    old_revno: Union[int, property]
    new_revno: Union[int, property]
    old_revid: RevisionID
    new_revid: RevisionID
    source_branch: Branch
    master_branch: Branch
    local_branch: Optional[Branch]
    target_branch: Branch
    tag_conflicts: List["TagConflict"]
    tag_updates: "TagUpdates"

    def report(self, to_file: TextIO) -> None:
        tag_conflicts = getattr(self, "tag_conflicts", None)
        tag_updates = getattr(self, "tag_updates", None)
        if not is_quiet():
            if self.old_revid != self.new_revid:
                to_file.write('Now on revision %d.\n' % self.new_revno)
            if tag_updates:
                to_file.write('%d tag(s) updated.\n' % len(tag_updates))
            if self.old_revid == self.new_revid and not tag_updates:
                if not tag_conflicts:
                    to_file.write('No revisions or tags to pull.\n')
                else:
                    to_file.write('No revisions to pull.\n')
        self._show_tag_conficts(to_file)


class BranchPushResult(_Result):
    """Result of a Branch.push operation.

    Attributes:
      old_revno: Revision number (eg 10) of the target before push.
      new_revno: Revision number (eg 12) of the target after push.
      old_revid: Tip revision id (eg joe@foo.com-1234234-aoeua34) of target
        before the push.
      new_revid: Tip revision id (eg joe@foo.com-5676566-boa234a) of target
        after the push.
      source_branch: Source branch object that the push was from. This is
        read locked, and generally is a local (and thus low latency) branch.
      master_branch: If target is a bound branch, the master branch of
        target, or target itself. Always write locked.
      target_branch: The direct Branch where data is being sent (write
        locked).
      local_branch: If the target is a bound branch this will be the
        target, otherwise it will be None.
    """

    old_revno: int
    new_revno: int
    old_revid: RevisionID
    new_revid: RevisionID
    source_branch: Branch
    master_branch: Branch
    target_branch: Branch
    local_branch: Optional[Branch]

    def report(self, to_file: TextIO) -> None:
        from breezy.i18n import gettext, ngettext

        # TODO: This function gets passed a to_file, but then
        # ignores it and calls note() instead. This is also
        # inconsistent with PullResult(), which writes to stdout.
        # -- JRV20110901, bug #838853
        tag_conflicts = getattr(self, "tag_conflicts", None)
        tag_updates = getattr(self, "tag_updates", None)
        if not is_quiet():
            if self.old_revid != self.new_revid:
                if self.new_revno is not None:
                    note(gettext('Pushed up to revision %d.'),
                         self.new_revno)
                else:
                    note(gettext('Pushed up to revision id %s.'),
                         self.new_revid.decode('utf-8'))
            if tag_updates:
                note(ngettext('%d tag updated.', '%d tags updated.',
                              len(tag_updates)) % len(tag_updates))
            if self.old_revid == self.new_revid and not tag_updates:
                if not tag_conflicts:
                    note(gettext('No new revisions or tags to push.'))
                else:
                    note(gettext('No new revisions to push.'))
        self._show_tag_conficts(to_file)


class BranchCheckResult:
    """Results of checking branch consistency.

    See `Branch.check`
    """

    def __init__(self, branch):
        self.branch = branch
        self.errors = []

    def report_results(self, verbose: bool) -> None:
        """Report the check results via trace.note.

        Args:
          verbose: Requests more detailed display of what was checked,
            if any.
        """
        from breezy.i18n import gettext, ngettext
        note(gettext('checked branch {0} format {1}').format(
            self.branch.user_url, self.branch._format))
        for error in self.errors:
            note(gettext('found error:%s'), error)


class InterBranch(InterObject[Branch]):
    """This class represents operations taking place between two branches.

    Its instances have methods like pull() and push() and contain
    references to the source and target repositories these operations
    can be carried out on.
    """

    _optimisers = []
    """The available optimised InterBranch types."""

    @classmethod
    def _get_branch_formats_to_test(klass):
        """Return an iterable of format tuples for testing.

        Returns: An iterable of (from_format, to_format) to use when testing
            this InterBranch class. Each InterBranch class should define this
            method itself.
        """
        raise NotImplementedError(klass._get_branch_formats_to_test)

    def pull(self, overwrite: bool = False,
             stop_revision: Optional[RevisionID] = None,
             possible_transports: Optional[List[Transport]] = None,
             local: bool = False, tag_selector=None) -> PullResult:
        """Mirror source into target branch.

        The target branch is considered to be 'local', having low latency.

        Returns: PullResult instance
        """
        raise NotImplementedError(self.pull)

    def push(self, overwrite: bool = False, stop_revision: Optional[RevisionID] = None,
             lossy: bool = False, _override_hook_source_branch: Optional[Branch] = None,
             tag_selector=None):
        """Mirror the source branch into the target branch.

        The source branch is considered to be 'local', having low latency.
        """
        raise NotImplementedError(self.push)

    def copy_content_into(self, revision_id=None, tag_selector=None):
        """Copy the content of source into target

        Args:
          revision_id:
            if not None, the revision history in the new branch will
            be truncated to end with revision_id.
          tag_selector: Optional callback that can decide
            to copy or not copy tags.
        """
        raise NotImplementedError(self.copy_content_into)

    def fetch(self, stop_revision: Optional[RevisionID] = None,
              limit: Optional[int] = None,
              lossy: bool = False) -> repository.FetchResult:
        """Fetch revisions.

        Args:
          stop_revision: Last revision to fetch
          limit: Optional rough limit of revisions to fetch
        Returns: FetchResult object
        """
        raise NotImplementedError(self.fetch)

    def update_references(self) -> None:
        """Import reference information from source to target.
        """
        raise NotImplementedError(self.update_references)

    @classmethod
    def get(self, source: Branch, target: Branch) -> "InterBranch":
        return cast(InterBranch, super().get(source, target))


def _fix_overwrite_type(overwrite):
    if isinstance(overwrite, bool):
        if overwrite:
            return ["history", "tags"]
        else:
            return []
    return overwrite


class GenericInterBranch(InterBranch):
    """InterBranch implementation that uses public Branch functions."""

    @classmethod
    def is_compatible(klass, source, target):
        # GenericBranch uses the public API, so always compatible
        return True

    @classmethod
    def _get_branch_formats_to_test(klass):
        return [(format_registry.get_default(), format_registry.get_default())]

    @classmethod
    def unwrap_format(klass, format):
        if isinstance(format, remote.RemoteBranchFormat):
            format._ensure_real()
            return format._custom_format
        return format

    def copy_content_into(self, revision_id=None, tag_selector=None):
        """Copy the content of source into target

        revision_id: if not None, the revision history in the new branch will
                     be truncated to end with revision_id.
        """
        with self.source.lock_read(), self.target.lock_write():
            self.source._synchronize_history(self.target, revision_id)
            self.update_references()
            try:
                parent = self.source.get_parent()
            except errors.InaccessibleParent as e:
                mutter('parent was not accessible to copy: %s', str(e))
            else:
                if parent:
                    self.target.set_parent(parent)
            if self.source._push_should_merge_tags():
                self.source.tags.merge_to(self.target.tags, selector=tag_selector)

    def fetch(self, stop_revision=None, limit=None, lossy=False):
        if self.target.base == self.source.base:
            return (0, [])
        with self.source.lock_read(), self.target.lock_write():
            fetch_spec_factory = fetch.FetchSpecFactory()
            fetch_spec_factory.source_branch = self.source
            fetch_spec_factory.source_branch_stop_revision_id = stop_revision
            fetch_spec_factory.source_repo = self.source.repository
            fetch_spec_factory.target_repo = self.target.repository
            fetch_spec_factory.target_repo_kind = (
                fetch.TargetRepoKinds.PREEXISTING)
            fetch_spec_factory.limit = limit
            fetch_spec = fetch_spec_factory.make_fetch_spec()
            return self.target.repository.fetch(
                self.source.repository,
                lossy=lossy,
                fetch_spec=fetch_spec)

    def _update_revisions(self, stop_revision=None, overwrite=False,
                          graph=None):
        with self.source.lock_read(), self.target.lock_write():
            other_revno, other_last_revision = self.source.last_revision_info()
            stop_revno = None  # unknown
            if stop_revision is None:
                stop_revision = other_last_revision
                if _mod_revision.is_null(stop_revision):
                    # if there are no commits, we're done.
                    return
                stop_revno = other_revno

            # what's the current last revision, before we fetch [and change it
            # possibly]
            last_rev = self.target.last_revision()
            # we fetch here so that we don't process data twice in the common
            # case of having something to pull, and so that the check for
            # already merged can operate on the just fetched graph, which will
            # be cached in memory.
            self.fetch(stop_revision=stop_revision)
            # Check to see if one is an ancestor of the other
            if not overwrite:
                if graph is None:
                    graph = self.target.repository.get_graph()
                if self.target._check_if_descendant_or_diverged(
                        stop_revision, last_rev, graph, self.source):
                    # stop_revision is a descendant of last_rev, but we aren't
                    # overwriting, so we're done.
                    return
            if stop_revno is None:
                if graph is None:
                    graph = self.target.repository.get_graph()
                this_revno, this_last_revision = \
                    self.target.last_revision_info()
                stop_revno = graph.find_distance_to_null(
                    stop_revision, [(other_last_revision, other_revno),
                                    (this_last_revision, this_revno)])
            self.target.set_last_revision_info(stop_revno, stop_revision)

    def pull(self, overwrite=False, stop_revision=None,
             possible_transports=None, run_hooks=True,
             _override_hook_target=None, local=False,
             tag_selector=None):
        """Pull from source into self, updating my master if any.

        Args:
          run_hooks: Private parameter - if false, this branch
            is being called because it's the master of the primary branch,
            so it should not run its hooks.
        """
        with contextlib.ExitStack() as exit_stack:
            exit_stack.enter_context(self.target.lock_write())
            bound_location = self.target.get_bound_location()
            if local and not bound_location:
                raise errors.LocalRequiresBoundBranch()
            master_branch = None
            source_is_master = False
            if bound_location:
                # bound_location comes from a config file, some care has to be
                # taken to relate it to source.user_url
                normalized = urlutils.normalize_url(bound_location)
                try:
                    relpath = self.source.user_transport.relpath(normalized)
                    source_is_master = (relpath == '')
                except (errors.PathNotChild, urlutils.InvalidURL):
                    source_is_master = False
            if not local and bound_location and not source_is_master:
                # not pulling from master, so we need to update master.
                master_branch = self.target.get_master_branch(
                    possible_transports)
                exit_stack.enter_context(master_branch.lock_write())
            if master_branch:
                # pull from source into master.
                master_branch.pull(
                    self.source, overwrite=overwrite, stop_revision=stop_revision,
                    run_hooks=False,
                    tag_selector=tag_selector)
            return self._pull(
                overwrite, stop_revision, _hook_master=master_branch,
                run_hooks=run_hooks,
                _override_hook_target=_override_hook_target,
                merge_tags_to_master=not source_is_master,
                tag_selector=tag_selector)

    def push(self, overwrite=False, stop_revision=None, lossy=False,
             _override_hook_source_branch=None, tag_selector=None):
        """See InterBranch.push.

        This is the basic concrete implementation of push()

        Args:
          _override_hook_source_branch: If specified, run the hooks
            passing this Branch as the source, rather than self.  This is for
            use of RemoteBranch, where push is delegated to the underlying
            vfs-based Branch.
        """
        if lossy:
            raise errors.LossyPushToSameVCS(self.source, self.target)
        # TODO: Public option to disable running hooks - should be trivial but
        # needs tests.

        def _run_hooks():
            if _override_hook_source_branch:
                result.source_branch = _override_hook_source_branch
            for hook in Branch.hooks['post_push']:
                hook(result)

        with self.source.lock_read(), self.target.lock_write():
            bound_location = self.target.get_bound_location()
            if bound_location and self.target.base != bound_location:
                # there is a master branch.
                #
                # XXX: Why the second check?  Is it even supported for a branch
                # to be bound to itself? -- mbp 20070507
                master_branch = self.target.get_master_branch()
                with master_branch.lock_write():
                    # push into the master from the source branch.
                    master_inter = InterBranch.get(self.source, master_branch)
                    master_inter._basic_push(
                        overwrite, stop_revision, tag_selector=tag_selector)
                    # and push into the target branch from the source. Note
                    # that we push from the source branch again, because it's
                    # considered the highest bandwidth repository.
                    result = self._basic_push(
                        overwrite, stop_revision, tag_selector=tag_selector)
                    result.master_branch = master_branch
                    result.local_branch = self.target
                    _run_hooks()
            else:
                master_branch = None
                # no master branch
                result = self._basic_push(
                    overwrite, stop_revision, tag_selector=tag_selector)
                # TODO: Why set master_branch and local_branch if there's no
                # binding?  Maybe cleaner to just leave them unset? -- mbp
                # 20070504
                result.master_branch = self.target
                result.local_branch = None
                _run_hooks()
            return result

    def _basic_push(self, overwrite, stop_revision, tag_selector=None):
        """Basic implementation of push without bound branches or hooks.

        Must be called with source read locked and target write locked.
        """
        result = BranchPushResult()
        result.source_branch = self.source
        result.target_branch = self.target
        result.old_revno, result.old_revid = self.target.last_revision_info()
        overwrite = _fix_overwrite_type(overwrite)
        if result.old_revid != stop_revision:
            # We assume that during 'push' this repository is closer than
            # the target.
            graph = self.source.repository.get_graph(self.target.repository)
            self._update_revisions(
                stop_revision, overwrite=("history" in overwrite), graph=graph)
        if self.source._push_should_merge_tags():
            result.tag_updates, result.tag_conflicts = (
                self.source.tags.merge_to(
                    self.target.tags, "tags" in overwrite, selector=tag_selector))
        self.update_references()
        result.new_revno, result.new_revid = self.target.last_revision_info()
        return result

    def _pull(self, overwrite=False, stop_revision=None,
              possible_transports=None, _hook_master=None, run_hooks=True,
              _override_hook_target=None, local=False,
              merge_tags_to_master=True, tag_selector=None):
        """See Branch.pull.

        This function is the core worker, used by GenericInterBranch.pull to
        avoid duplication when pulling source->master and source->local.

        Args:
          _hook_master: Private parameter - set the branch to
            be supplied as the master to pull hooks.
          run_hooks: Private parameter - if false, this branch
            is being called because it's the master of the primary branch,
            so it should not run its hooks.
            is being called because it's the master of the primary branch,
            so it should not run its hooks.
          _override_hook_target: Private parameter - set the branch to be
            supplied as the target_branch to pull hooks.
          local: Only update the local branch, and not the bound branch.
        """
        # This type of branch can't be bound.
        if local:
            raise errors.LocalRequiresBoundBranch()
        result = PullResult()
        result.source_branch = self.source
        if _override_hook_target is None:
            result.target_branch = self.target
        else:
            result.target_branch = _override_hook_target
        with self.source.lock_read():
            # We assume that during 'pull' the target repository is closer than
            # the source one.
            graph = self.target.repository.get_graph(self.source.repository)
            # TODO: Branch formats should have a flag that indicates
            # that revno's are expensive, and pull() should honor that flag.
            # -- JRV20090506
            result.old_revno, result.old_revid = \
                self.target.last_revision_info()
            overwrite = _fix_overwrite_type(overwrite)
            self._update_revisions(
                stop_revision, overwrite=("history" in overwrite), graph=graph)
            # TODO: The old revid should be specified when merging tags,
            # so a tags implementation that versions tags can only
            # pull in the most recent changes. -- JRV20090506
            result.tag_updates, result.tag_conflicts = (
                self.source.tags.merge_to(
                    self.target.tags, "tags" in overwrite,
                    ignore_master=not merge_tags_to_master,
                    selector=tag_selector))
            self.update_references()
            result.new_revno, result.new_revid = (
                self.target.last_revision_info())
            if _hook_master:
                result.master_branch = _hook_master
                result.local_branch = result.target_branch
            else:
                result.master_branch = result.target_branch
                result.local_branch = None
            if run_hooks:
                for hook in Branch.hooks['post_pull']:
                    hook(result)
            return result

    def update_references(self):
        if not getattr(self.source._format, 'supports_reference_locations', False):
            return
        reference_dict = self.source._get_all_reference_info()
        if len(reference_dict) == 0:
            return
        old_base = self.source.base
        new_base = self.target.base
        target_reference_dict = self.target._get_all_reference_info()
        for tree_path, (branch_location, file_id) in reference_dict.items():
            try:
                branch_location = urlutils.rebase_url(branch_location,
                                                      old_base, new_base)
            except urlutils.InvalidRebaseURLs:
                # Fall back to absolute URL
                branch_location = urlutils.join(old_base, branch_location)
            target_reference_dict.setdefault(
                tree_path, (branch_location, file_id))
        self.target._set_all_reference_info(target_reference_dict)


InterBranch.register_optimiser(GenericInterBranch)



# Copyright (C) 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for add_signature_text on a repository with external references."""

from breezy import errors
from breezy.repository import WriteGroup
from breezy.tests.per_repository_reference import \
    TestCaseWithExternalReferenceRepository


class TestAddSignatureText(TestCaseWithExternalReferenceRepository):

    def test_add_signature_text_goes_to_repo(self):
        # adding a signature only writes to the repository add_signature_text
        # is called on.
        tree = self.make_branch_and_tree('sample')
        revid = tree.commit('one')
        inv = tree.branch.repository.get_inventory(revid)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        base = self.make_repository('base')
        repo = self.make_referring('referring', base)
        with repo.lock_write(), WriteGroup(repo):
            rev = tree.branch.repository.get_revision(revid)
            repo.texts.add_lines((inv.root.file_id, revid), [], [])
            repo.add_revision(revid, rev, inv=inv)
            repo.add_signature_text(revid, b"text")
        repo.get_signature_text(revid)
        self.assertRaises(errors.NoSuchRevision, base.get_signature_text,
                          revid)



# Copyright (C) 2006 Canonical Ltd
# Authors:  Robert Collins <robert.collins@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for the MemoryTree class."""

from .. import errors, transport
from ..memorytree import MemoryTree
from ..treebuilder import TreeBuilder
from . import TestCaseWithTransport


class TestMemoryTree(TestCaseWithTransport):

    def test_create_on_branch(self):
        """Creating a mutable tree on a trivial branch works."""
        branch = self.make_branch('branch')
        tree = MemoryTree.create_on_branch(branch)
        self.assertEqual(branch.controldir, tree.controldir)
        self.assertEqual(branch, tree.branch)
        self.assertEqual([], tree.get_parent_ids())

    def test_create_on_branch_with_content(self):
        """Creating a mutable tree on a non-trivial branch works."""
        branch = self.make_branch('branch')
        tree = MemoryTree.create_on_branch(branch)
        # build some content
        tree.lock_write()
        builder = TreeBuilder()
        builder.start_tree(tree)
        builder.build(['foo'])
        builder.finish_tree()
        rev_id = tree.commit('first post')
        tree.unlock()
        tree = MemoryTree.create_on_branch(branch)
        with tree.lock_read():
            self.assertEqual([rev_id], tree.get_parent_ids())
            with tree.get_file('foo') as f:
                self.assertEqual(b'contents of foo\n', f.read())

    def test_get_root_id(self):
        branch = self.make_branch('branch')
        tree = MemoryTree.create_on_branch(branch)
        with tree.lock_write():
            tree.add([''])
            self.assertIsNot(None, tree.path2id(''))

    def test_lock_tree_write(self):
        """Check we can lock_tree_write and unlock MemoryTrees."""
        branch = self.make_branch('branch')
        tree = MemoryTree.create_on_branch(branch)
        tree.lock_tree_write()
        tree.unlock()

    def test_lock_tree_write_after_read_fails(self):
        """Check that we error when trying to upgrade a read lock to write."""
        branch = self.make_branch('branch')
        tree = MemoryTree.create_on_branch(branch)
        with tree.lock_read():
            self.assertRaises(errors.ReadOnlyError, tree.lock_tree_write)

    def test_lock_write(self):
        """Check we can lock_write and unlock MemoryTrees."""
        branch = self.make_branch('branch')
        tree = MemoryTree.create_on_branch(branch)
        tree.lock_write()
        tree.unlock()

    def test_lock_write_after_read_fails(self):
        """Check that we error when trying to upgrade a read lock to write."""
        branch = self.make_branch('branch')
        tree = MemoryTree.create_on_branch(branch)
        with tree.lock_read():
            self.assertRaises(errors.ReadOnlyError, tree.lock_write)

    def test_add_with_kind(self):
        branch = self.make_branch('branch')
        tree = MemoryTree.create_on_branch(branch)
        with tree.lock_write():
            tree.add(['', 'afile', 'adir'], ['directory', 'file', 'directory'])
            self.assertEqual('afile', tree.id2path(tree.path2id('afile')))
            self.assertEqual('adir', tree.id2path(tree.path2id('adir')))
            self.assertFalse(tree.has_filename('afile'))
            self.assertFalse(tree.has_filename('adir'))

    def test_put_new_file(self):
        branch = self.make_branch('branch')
        tree = MemoryTree.create_on_branch(branch)
        with tree.lock_write():
            tree.add(['', 'foo'], ids=[b'root-id', b'foo-id'],
                     kinds=['directory', 'file'])
            tree.put_file_bytes_non_atomic('foo', b'barshoom')
            with tree.get_file('foo') as f:
                self.assertEqual(b'barshoom', f.read())

    def test_put_existing_file(self):
        branch = self.make_branch('branch')
        tree = MemoryTree.create_on_branch(branch)
        with tree.lock_write():
            tree.add(['', 'foo'], ids=[b'root-id', b'foo-id'],
                     kinds=['directory', 'file'])
            tree.put_file_bytes_non_atomic('foo', b'first-content')
            tree.put_file_bytes_non_atomic('foo', b'barshoom')
            self.assertEqual(b'barshoom', tree.get_file('foo').read())

    def test_add_in_subdir(self):
        branch = self.make_branch('branch')
        tree = MemoryTree.create_on_branch(branch)
        with tree.lock_write():
            tree.add([''], ['directory'], [b'root-id'])
            # Unfortunately, the only way to 'mkdir' is to call 'tree.mkdir', but
            # that *always* adds the directory as well. So if you want to create a
            # file in a subdirectory, you have to split out the 'mkdir()' calls
            # from the add and put_file_bytes_non_atomic calls. :(
            tree.mkdir('adir', b'dir-id')
            tree.add(['adir/afile'], ids=[b'file-id'], kinds=['file'])
            self.assertEqual('adir/afile', tree.id2path(b'file-id'))
            self.assertEqual('adir', tree.id2path(b'dir-id'))
            tree.put_file_bytes_non_atomic('adir/afile', b'barshoom')

    def test_add_symlink(self):
        branch = self.make_branch('branch')
        tree = MemoryTree.create_on_branch(branch)
        with tree.lock_write():
            tree._file_transport.symlink('bar', 'foo')
            tree.add(['', 'foo'])
            self.assertEqual('symlink', tree.kind('foo'))
            self.assertEqual('bar', tree.get_symlink_target('foo'))

    def test_commit_trivial(self):
        """Smoke test for commit on a MemoryTree.

        Becamse of commits design and layering, if this works, all commit
        logic should work quite reliably.
        """
        branch = self.make_branch('branch')
        tree = MemoryTree.create_on_branch(branch)
        with tree.lock_write():
            tree.add(['', 'foo'], ids=[b'root-id', b'foo-id'],
                     kinds=['directory', 'file'])
            tree.put_file_bytes_non_atomic('foo', b'barshoom')
            revision_id = tree.commit('message baby')
            # the parents list for the tree should have changed.
            self.assertEqual([revision_id], tree.get_parent_ids())
        # and we should have a revision that is accessible outside the tree lock
        revtree = tree.branch.repository.revision_tree(revision_id)
        with revtree.lock_read(), revtree.get_file('foo') as f:
            self.assertEqual(b'barshoom', f.read())

    def test_unversion(self):
        """Some test for unversion of a memory tree."""
        branch = self.make_branch('branch')
        tree = MemoryTree.create_on_branch(branch)
        with tree.lock_write():
            tree.add(['', 'foo'], ids=[b'root-id', b'foo-id'],
                     kinds=['directory', 'file'])
            tree.unversion(['foo'])
            self.assertFalse(tree.is_versioned('foo'))
            self.assertRaises(errors.NoSuchId, tree.id2path, b'foo-id')

    def test_last_revision(self):
        """There should be a last revision method we can call."""
        tree = self.make_branch_and_memory_tree('branch')
        with tree.lock_write():
            tree.add('')
            rev_id = tree.commit('first post')
        self.assertEqual(rev_id, tree.last_revision())

    def test_rename_file(self):
        tree = self.make_branch_and_memory_tree('branch')
        tree.lock_write()
        self.addCleanup(tree.unlock)
        tree.add(['', 'foo'], ['directory', 'file'], [b'root-id', b'foo-id'])
        tree.put_file_bytes_non_atomic('foo', b'content\n')
        tree.commit('one', rev_id=b'rev-one')
        tree.rename_one('foo', 'bar')
        self.assertEqual('bar', tree.id2path(b'foo-id'))
        self.assertEqual(b'content\n', tree._file_transport.get_bytes('bar'))
        self.assertRaises(transport.NoSuchFile,
                          tree._file_transport.get_bytes, 'foo')
        tree.commit('two', rev_id=b'rev-two')
        self.assertEqual(b'content\n', tree._file_transport.get_bytes('bar'))
        self.assertRaises(transport.NoSuchFile,
                          tree._file_transport.get_bytes, 'foo')

        rev_tree2 = tree.branch.repository.revision_tree(b'rev-two')
        self.assertEqual('bar', rev_tree2.id2path(b'foo-id'))
        self.assertEqual(b'content\n', rev_tree2.get_file_text('bar'))

    def test_rename_file_to_subdir(self):
        tree = self.make_branch_and_memory_tree('branch')
        tree.lock_write()
        self.addCleanup(tree.unlock)
        tree.add('')
        tree.mkdir('subdir', b'subdir-id')
        tree.add('foo', 'file', b'foo-id')
        tree.put_file_bytes_non_atomic('foo', b'content\n')
        tree.commit('one', rev_id=b'rev-one')

        tree.rename_one('foo', 'subdir/bar')
        self.assertEqual('subdir/bar', tree.id2path(b'foo-id'))
        self.assertEqual(b'content\n',
                         tree._file_transport.get_bytes('subdir/bar'))
        tree.commit('two', rev_id=b'rev-two')
        rev_tree2 = tree.branch.repository.revision_tree(b'rev-two')
        self.assertEqual('subdir/bar', rev_tree2.id2path(b'foo-id'))



# Copyright (C) 2010-2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for pushing revisions from Bazaar into Git."""

from ...controldir import format_registry
from ...repository import InterRepository
from ...tests import TestCaseWithTransport
from ..interrepo import InterToGitRepository
from ..mapping import BzrGitMappingExperimental, BzrGitMappingv1


class InterToGitRepositoryTests(TestCaseWithTransport):

    def setUp(self):
        super().setUp()
        self.git_repo = self.make_repository("git",
                                             format=format_registry.make_controldir("git"))
        self.bzr_repo = self.make_repository("bzr", shared=True)

    def _get_interrepo(self, mapping=None):
        self.bzr_repo.lock_read()
        self.addCleanup(self.bzr_repo.unlock)
        interrepo = InterRepository.get(self.bzr_repo, self.git_repo)
        if mapping is not None:
            interrepo.mapping = mapping
        return interrepo

    def test_instance(self):
        self.assertIsInstance(self._get_interrepo(), InterToGitRepository)

    def test_pointless_fetch_refs_old_mapping(self):
        interrepo = self._get_interrepo(mapping=BzrGitMappingv1())
        interrepo.fetch_refs(lambda x: {}, lossy=False)

    def test_pointless_fetch_refs(self):
        interrepo = self._get_interrepo(mapping=BzrGitMappingExperimental())
        revidmap, old_refs, new_refs = interrepo.fetch_refs(
            lambda x: {}, lossy=False)
        self.assertEqual(old_refs, {b'HEAD': (
            b'ref: refs/heads/master', None)})
        self.assertEqual(new_refs, {})

    def test_pointless_lossy_fetch_refs(self):
        revidmap, old_refs, new_refs = self._get_interrepo(
        ).fetch_refs(lambda x: {}, lossy=True)
        self.assertEqual(old_refs, {b'HEAD': (
            b'ref: refs/heads/master', None)})
        self.assertEqual(new_refs, {})
        self.assertEqual(revidmap, {})

    def test_pointless_missing_revisions(self):
        interrepo = self._get_interrepo()
        interrepo.source_store.lock_read()
        self.addCleanup(interrepo.source_store.unlock)
        self.assertEqual([], list(interrepo.missing_revisions([])))

    def test_missing_revisions_unknown_stop_rev(self):
        interrepo = self._get_interrepo()
        interrepo.source_store.lock_read()
        self.addCleanup(interrepo.source_store.unlock)
        self.assertEqual([],
                         list(interrepo.missing_revisions([(None, b"unknown")])))

    def test_odd_rename(self):
        # Add initial revision to bzr branch.
        branch = self.bzr_repo.controldir.create_branch()
        tree = branch.controldir.create_workingtree()
        self.build_tree(["bzr/bar/", "bzr/bar/foobar"])
        tree.add(["bar", "bar/foobar"])
        tree.commit("initial")

        # Add new directory and perform move in bzr branch.
        self.build_tree(["bzr/baz/"])
        tree.add(["baz"])
        tree.rename_one("bar", "baz/IrcDotNet")
        last_revid = tree.commit("rename")

        # Push bzr branch to git branch.
        def decide(x):
            return {b"refs/heads/master": (None, last_revid)}
        interrepo = self._get_interrepo()
        revidmap, old_refs, new_refs = interrepo.fetch_refs(decide, lossy=True)
        gitid = revidmap[last_revid][0]
        store = self.git_repo._git.object_store
        commit = store[gitid]
        tree = store[commit.tree]
        tree.check()
        self.assertIn(b"baz", tree, repr(tree.items()))
        self.assertIn(tree[b"baz"][1], store)
        baz = store[tree[b"baz"][1]]
        baz.check()
        ircdotnet = store[baz[b"IrcDotNet"][1]]
        ircdotnet.check()
        foobar = store[ircdotnet[b"foobar"][1]]
        foobar.check()



# Copyright (C) 2007-2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for win32utils."""

import os
from typing import List

from .. import osutils, tests, win32utils
from ..win32utils import get_app_path, glob_expand
from . import TestCase, TestCaseInTempDir, TestSkipped, features
from .features import backslashdir_feature

Win32RegistryFeature = features.ModuleAvailableFeature('_winreg')


class TestWin32UtilsGlobExpand(TestCaseInTempDir):

    _test_needs_features: List[features.Feature] = []

    def test_empty_tree(self):
        self.build_tree([])
        self._run_testset([
            [['a'], ['a']],
            [['?'], ['?']],
            [['*'], ['*']],
            [['a', 'a'], ['a', 'a']]])

    def build_ascii_tree(self):
        self.build_tree(['a', 'a1', 'a2', 'a11', 'a.1',
                         'b', 'b1', 'b2', 'b3',
                         'c/', 'c/c1', 'c/c2',
                         'd/', 'd/d1', 'd/d2', 'd/e/', 'd/e/e1'])

    def build_unicode_tree(self):
        self.requireFeature(features.UnicodeFilenameFeature)
        self.build_tree(['\u1234', '\u1234\u1234', '\u1235/',
                         '\u1235/\u1235'])

    def test_tree_ascii(self):
        """Checks the glob expansion and path separation char
        normalization"""
        self.build_ascii_tree()
        self._run_testset([
            # no wildcards
            [['a'], ['a']],
            [['a', 'a'], ['a', 'a']],

            [['d'], ['d']],
            [['d/'], ['d/']],

            # wildcards
            [['a*'], ['a', 'a1', 'a2', 'a11', 'a.1']],
            [['?'], ['a', 'b', 'c', 'd']],
            [['a?'], ['a1', 'a2']],
            [['a??'], ['a11', 'a.1']],
            [['b[1-2]'], ['b1', 'b2']],

            [['d/*'], ['d/d1', 'd/d2', 'd/e']],
            [['?/*'], ['c/c1', 'c/c2', 'd/d1', 'd/d2', 'd/e']],
            [['*/*'], ['c/c1', 'c/c2', 'd/d1', 'd/d2', 'd/e']],
            [['*/'], ['c/', 'd/']],
            ])

    def test_backslash_globbing(self):
        self.requireFeature(backslashdir_feature)
        self.build_ascii_tree()
        self._run_testset([
            [['d\\'], ['d/']],
            [['d\\*'], ['d/d1', 'd/d2', 'd/e']],
            [['?\\*'], ['c/c1', 'c/c2', 'd/d1', 'd/d2', 'd/e']],
            [['*\\*'], ['c/c1', 'c/c2', 'd/d1', 'd/d2', 'd/e']],
            [['*\\'], ['c/', 'd/']],
            ])

    def test_case_insensitive_globbing(self):
        if os.path.normcase("AbC") == "AbC":
            self.skipTest("Test requires case insensitive globbing function")
        self.build_ascii_tree()
        self._run_testset([
            [['A'], ['A']],
            [['A?'], ['a1', 'a2']],
            ])

    def test_tree_unicode(self):
        """Checks behaviour with non-ascii filenames"""
        self.build_unicode_tree()
        self._run_testset([
            # no wildcards
            [['\u1234'], ['\u1234']],
            [['\u1235'], ['\u1235']],

            [['\u1235/'], ['\u1235/']],
            [['\u1235/\u1235'], ['\u1235/\u1235']],

            # wildcards
            [['?'], ['\u1234', '\u1235']],
            [['*'], ['\u1234', '\u1234\u1234', '\u1235']],
            [['\u1234*'], ['\u1234', '\u1234\u1234']],

            [['\u1235/?'], ['\u1235/\u1235']],
            [['\u1235/*'], ['\u1235/\u1235']],
            [['?/'], ['\u1235/']],
            [['*/'], ['\u1235/']],
            [['?/?'], ['\u1235/\u1235']],
            [['*/*'], ['\u1235/\u1235']],
            ])

    def test_unicode_backslashes(self):
        self.requireFeature(backslashdir_feature)
        self.build_unicode_tree()
        self._run_testset([
            # no wildcards
            [['\u1235\\'], ['\u1235/']],
            [['\u1235\\\u1235'], ['\u1235/\u1235']],
            [['\u1235\\?'], ['\u1235/\u1235']],
            [['\u1235\\*'], ['\u1235/\u1235']],
            [['?\\'], ['\u1235/']],
            [['*\\'], ['\u1235/']],
            [['?\\?'], ['\u1235/\u1235']],
            [['*\\*'], ['\u1235/\u1235']],
            ])

    def _run_testset(self, testset):
        for pattern, expected in testset:
            result = glob_expand(pattern)
            expected.sort()
            result.sort()
            self.assertEqual(expected, result, 'pattern %s' % pattern)


class TestAppPaths(TestCase):

    _test_needs_features = [Win32RegistryFeature]

    def test_iexplore(self):
        # typical windows users should have IE installed
        for a in ('iexplore', 'iexplore.exe'):
            p = get_app_path(a)
            d, b = os.path.split(p)
            self.assertEqual('iexplore.exe', b.lower())
            self.assertNotEqual('', d)

    def test_wordpad(self):
        # typical windows users should have wordpad in the system
        # but there is problem: its path has the format REG_EXPAND_SZ
        # so naive attempt to get the path is not working
        self.requireFeature(Win32ApiFeature)
        for a in ('wordpad', 'wordpad.exe'):
            p = get_app_path(a)
            d, b = os.path.split(p)
            self.assertEqual('wordpad.exe', b.lower())
            self.assertNotEqual('', d)

    def test_not_existing(self):
        p = get_app_path('not-existing')
        self.assertEqual('not-existing', p)


class TestLocations(TestCase):

    _test_needs_features = [features.win32_feature]

    def assertPathsEqual(self, p1, p2):
        # TODO: The env var values in particular might return the "short"
        # version (ie, "C:\DOCUME~1\...").  Its even possible the returned
        # values will differ only by case - handle these situations as we
        # come across them.
        self.assertEqual(p1, p2)

    def test_appdata_not_using_environment(self):
        # Test that we aren't falling back to the environment
        first = win32utils.get_appdata_location()
        self.overrideEnv("APPDATA", None)
        self.assertPathsEqual(first, win32utils.get_appdata_location())

    def test_appdata_matches_environment(self):
        # Typically the APPDATA environment variable will match
        # get_appdata_location
        # XXX - See bug 262874, which asserts the correct encoding is 'mbcs',
        encoding = osutils.get_user_encoding()
        env_val = os.environ.get("APPDATA", None)
        if not env_val:
            raise TestSkipped("No APPDATA environment variable exists")
        self.assertPathsEqual(win32utils.get_appdata_location(),
                              env_val.decode(encoding))

    def test_local_appdata_not_using_environment(self):
        # Test that we aren't falling back to the environment
        first = win32utils.get_local_appdata_location()
        self.overrideEnv("LOCALAPPDATA", None)
        self.assertPathsEqual(first, win32utils.get_local_appdata_location())

    def test_local_appdata_matches_environment(self):
        # LOCALAPPDATA typically only exists on Vista, so we only attempt to
        # compare when it exists.
        lad = win32utils.get_local_appdata_location()
        env = os.environ.get("LOCALAPPDATA")
        if env:
            # XXX - See bug 262874, which asserts the correct encoding is
            # 'mbcs'
            encoding = osutils.get_user_encoding()
            self.assertPathsEqual(lad, env.decode(encoding))


class TestSetHidden(TestCaseInTempDir):

    _test_needs_features = [features.win32_feature]

    def test_unicode_dir(self):
        # we should handle unicode paths without errors
        self.requireFeature(features.UnicodeFilenameFeature)
        os.mkdir('\u1234')
        win32utils.set_file_attr_hidden('\u1234')

    def test_dot_bzr_in_unicode_dir(self):
        # we should not raise traceback if we try to set hidden attribute
        # on .bzr directory below unicode path
        self.requireFeature(features.UnicodeFilenameFeature)
        os.makedirs('\u1234\\.bzr')
        path = osutils.abspath('\u1234\\.bzr')
        win32utils.set_file_attr_hidden(path)


class Test_CommandLineToArgv(tests.TestCaseInTempDir):

    def assertCommandLine(self, expected, line, argv=None,
                          single_quotes_allowed=False):
        # Strictly speaking we should respect parameter order versus glob
        # expansions, but it's not really worth the effort here
        if argv is None:
            argv = [line]
        argv = win32utils._command_line_to_argv(
            line, argv, single_quotes_allowed=single_quotes_allowed)
        self.assertEqual(expected, sorted(argv))

    def test_glob_paths(self):
        self.build_tree(['a/', 'a/b.c', 'a/c.c', 'a/c.h'])
        self.assertCommandLine(['a/b.c', 'a/c.c'], 'a/*.c')
        self.build_tree(['b/', 'b/b.c', 'b/d.c', 'b/d.h'])
        self.assertCommandLine(['a/b.c', 'b/b.c'], '*/b.c')
        self.assertCommandLine(['a/b.c', 'a/c.c', 'b/b.c', 'b/d.c'],
                               '*/*.c')
        # Bash style, just pass through the argument if nothing matches
        self.assertCommandLine(['*/*.qqq'], '*/*.qqq')

    def test_quoted_globs(self):
        self.build_tree(['a/', 'a/b.c', 'a/c.c', 'a/c.h'])
        self.assertCommandLine(['a/*.c'], '"a/*.c"')
        self.assertCommandLine(["'a/*.c'"], "'a/*.c'")
        self.assertCommandLine(['a/*.c'], "'a/*.c'",
                               single_quotes_allowed=True)

    def test_slashes_changed(self):
        # Quoting doesn't change the supplied args
        self.assertCommandLine(['a\\*.c'], '"a\\*.c"')
        self.assertCommandLine(['a\\*.c'], "'a\\*.c'",
                               single_quotes_allowed=True)
        # Expands the glob, but nothing matches, swaps slashes
        self.assertCommandLine(['a/*.c'], 'a\\*.c')
        self.assertCommandLine(['a/?.c'], 'a\\?.c')
        # No glob, doesn't touch slashes
        self.assertCommandLine(['a\\foo.c'], 'a\\foo.c')

    def test_single_quote_support(self):
        self.assertCommandLine(["add", "let's-do-it.txt"],
                               "add let's-do-it.txt",
                               ["add", "let's-do-it.txt"])
        self.expectFailure("Using single quotes breaks trimming from argv",
                           self.assertCommandLine, ["add", "lets do it.txt"],
                           "add 'lets do it.txt'", [
                               "add", "'lets", "do", "it.txt'"],
                           single_quotes_allowed=True)

    def test_case_insensitive_globs(self):
        if os.path.normcase("AbC") == "AbC":
            self.skipTest("Test requires case insensitive globbing function")
        self.build_tree(['a/', 'a/b.c', 'a/c.c', 'a/c.h'])
        self.assertCommandLine(['A/b.c'], 'A/B*')

    def test_backslashes(self):
        self.requireFeature(backslashdir_feature)
        self.build_tree(['a/', 'a/b.c', 'a/c.c', 'a/c.h'])
        self.assertCommandLine(['a/b.c'], 'a\\b*')

    def test_with_pdb(self):
        """Check stripping Python arguments before bzr script per lp:587868"""
        self.assertCommandLine(["rocks"], "-m pdb rocks", ["rocks"])
        self.build_tree(['d/', 'd/f1', 'd/f2'])
        self.assertCommandLine(["rm", "x*"], "-m pdb rm x*", ["rm", "x*"])
        self.assertCommandLine(["add", "d/f1", "d/f2"], "-m pdb add d/*",
                               ["add", "d/*"])



# vim: expandtab

# Copyright (C) 2011-2018 Jelmer Vernooij <jelmer@jelmer.uk>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for the git remote helper."""

import os
import subprocess
import sys
from io import BytesIO

from dulwich.repo import Repo

from ...tests import TestCaseWithTransport
from ...tests.features import PathFeature
from ..git_remote_helper import RemoteHelper, fetch, open_local_dir
from ..object_store import get_object_store
from . import FastimportFeature


def map_to_git_sha1(dir, bzr_revid):
    object_store = get_object_store(dir.open_repository())
    with object_store.lock_read():
        return object_store._lookup_revision_sha1(bzr_revid)


git_remote_bzr_path = os.path.abspath(
    os.path.join(os.path.dirname(__file__), '..', 'git-remote-bzr'))
git_remote_bzr_feature = PathFeature(git_remote_bzr_path)


class OpenLocalDirTests(TestCaseWithTransport):

    def test_from_env_dir(self):
        self.make_branch_and_tree('bla', format='git')
        self.overrideEnv('GIT_DIR', os.path.join(self.test_dir, 'bla', '.git'))
        open_local_dir()

    def test_from_dir(self):
        self.make_branch_and_tree('.', format='git')
        open_local_dir()


class FetchTests(TestCaseWithTransport):

    def setUp(self):
        super().setUp()
        self.local_dir = self.make_branch_and_tree(
            'local', format='git').controldir
        self.remote_tree = self.make_branch_and_tree('remote')
        self.remote_dir = self.remote_tree.controldir
        self.shortname = 'bzr'

    def fetch(self, wants):
        outf = BytesIO()
        fetch(outf, wants, self.shortname, self.remote_dir, self.local_dir)
        return outf.getvalue()

    def test_no_wants(self):
        r = self.fetch([])
        self.assertEqual(b"\n", r)

    def test_simple(self):
        self.build_tree(['remote/foo'])
        self.remote_tree.add("foo")
        revid = self.remote_tree.commit("msg")
        git_sha1 = map_to_git_sha1(self.remote_dir, revid)
        out = self.fetch([(git_sha1, 'HEAD')])
        self.assertEqual(out, b"\n")
        r = Repo('local')
        self.assertTrue(git_sha1 in r.object_store)
        self.assertEqual({}, r.get_refs())


class ExecuteRemoteHelperTests(TestCaseWithTransport):

    def test_run(self):
        self.requireFeature(git_remote_bzr_feature)
        local_dir = self.make_branch_and_tree('local', format='git').controldir
        local_path = local_dir.control_transport.local_abspath('.')
        remote_tree = self.make_branch_and_tree('remote')
        remote_dir = remote_tree.controldir
        shortname = 'bzr'
        env = dict(os.environ)
        env['GIT_DIR'] = local_path
        env['PYTHONPATH'] = ':'.join(sys.path)
        p = subprocess.Popen(
            [sys.executable, git_remote_bzr_path, local_path, remote_dir.user_url],
            stdin=subprocess.PIPE, stdout=subprocess.PIPE,
            stderr=subprocess.PIPE, env=env)
        (out, err) = p.communicate(b'capabilities\n')
        lines = out.splitlines()
        self.assertIn(b'push', lines, "no 'push' in {!r}, error: {!r}".format(lines, err))
        self.assertEqual(
            b"git-remote-bzr is experimental and has not been optimized "
            b"for performance. Use 'brz fast-export' and 'git fast-import' "
            b"for large repositories.\n", err)


class RemoteHelperTests(TestCaseWithTransport):

    def setUp(self):
        super().setUp()
        self.local_dir = self.make_branch_and_tree(
            'local', format='git').controldir
        self.remote_tree = self.make_branch_and_tree('remote')
        self.remote_dir = self.remote_tree.controldir
        self.shortname = 'bzr'
        self.helper = RemoteHelper(
            self.local_dir, self.shortname, self.remote_dir)

    def test_capabilities(self):
        f = BytesIO()
        self.helper.cmd_capabilities(f, [])
        capabs = f.getvalue()
        base = b"fetch\noption\npush\n"
        self.assertTrue(capabs in (base + b"\n", base + b"import\nrefspec *:*\n\n"), capabs)

    def test_option(self):
        f = BytesIO()
        self.helper.cmd_option(f, [])
        self.assertEqual(b"unsupported\n", f.getvalue())

    def test_list_basic(self):
        f = BytesIO()
        self.helper.cmd_list(f, [])
        self.assertEqual(
            b'\n',
            f.getvalue())

    def test_import(self):
        self.requireFeature(FastimportFeature)
        self.build_tree_contents([("remote/afile", b"somecontent")])
        self.remote_tree.add(["afile"])
        self.remote_tree.commit(b"A commit message", timestamp=1330445983,
                                timezone=0, committer=b'Somebody <jrandom@example.com>')
        f = BytesIO()
        self.helper.cmd_import(f, ["import", "refs/heads/master"])
        self.assertEqual(
            b'reset refs/heads/master\n'
            b'commit refs/heads/master\n'
            b'mark :1\n'
            b'committer Somebody <jrandom@example.com> 1330445983 +0000\n'
            b'data 16\n'
            b'A commit message\n'
            b'M 644 inline afile\n'
            b'data 11\n'
            b'somecontent\n',
            f.getvalue())



# Copyright (C) 2006-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for branch.pull behaviour."""

from breezy import branch, controldir, errors, revision
from breezy.tests import TestNotApplicable, fixtures, per_branch


class TestPull(per_branch.TestCaseWithBranch):

    def test_pull_convergence_simple(self):
        # when revisions are pulled, the left-most accessible parents must
        # become the revision-history.
        parent = self.make_branch_and_tree('parent')
        p1 = parent.commit('1st post', allow_pointless=True)
        mine = parent.controldir.sprout('mine').open_workingtree()
        m1 = mine.commit('my change', allow_pointless=True)
        parent.merge_from_branch(mine.branch)
        p2 = parent.commit('merge my change')
        mine.pull(parent.branch)
        self.assertEqual(p2, mine.branch.last_revision())

    def test_pull_merged_indirect(self):
        # it should be possible to do a pull from one branch into another
        # when the tip of the target was merged into the source branch
        # via a third branch - so its buried in the ancestry and is not
        # directly accessible.
        parent = self.make_branch_and_tree('parent')
        p1 = parent.commit('1st post', allow_pointless=True)
        mine = parent.controldir.sprout('mine').open_workingtree()
        m1 = mine.commit('my change', allow_pointless=True)
        other = parent.controldir.sprout('other').open_workingtree()
        other.merge_from_branch(mine.branch)
        o2 = other.commit('merge my change')
        parent.merge_from_branch(other.branch)
        p2 = parent.commit('merge other')
        mine.pull(parent.branch)
        self.assertEqual(p2, mine.branch.last_revision())

    def test_pull_updates_checkout_and_master(self):
        """Pulling into a checkout updates the checkout and the master branch"""
        master_tree = self.make_branch_and_tree('master')
        rev1 = master_tree.commit('master')
        checkout = master_tree.branch.create_checkout('checkout')

        other = master_tree.branch.controldir.sprout(
            'other').open_workingtree()
        rev2 = other.commit('other commit')
        # now pull, which should update both checkout and master.
        checkout.branch.pull(other.branch)
        self.assertEqual(rev2, checkout.branch.last_revision())
        self.assertEqual(rev2, master_tree.branch.last_revision())

    def test_pull_local_updates_checkout_only(self):
        """Pulling --local into a checkout updates the checkout and not the
        master branch"""
        master_tree = self.make_branch_and_tree('master')
        rev1 = master_tree.commit('master')
        checkout = master_tree.branch.create_checkout('checkout')

        other = master_tree.branch.controldir.sprout(
            'other').open_workingtree()
        rev2 = other.commit('other commit')
        # now pull local, which should update checkout but not master.
        checkout.branch.pull(other.branch, local=True)
        self.assertEqual(rev2, checkout.branch.last_revision())
        self.assertEqual(rev1, master_tree.branch.last_revision())

    def test_pull_local_raises_LocalRequiresBoundBranch_on_unbound(self):
        """Pulling --local into a branch that is not bound should fail."""
        master_tree = self.make_branch_and_tree('branch')
        rev1 = master_tree.commit('master')

        other = master_tree.branch.controldir.sprout(
            'other').open_workingtree()
        rev2 = other.commit('other commit')
        # now pull --local, which should raise LocalRequiresBoundBranch error.
        self.assertRaises(errors.LocalRequiresBoundBranch,
                          master_tree.branch.pull, other.branch, local=True)
        self.assertEqual(rev1, master_tree.branch.last_revision())

    def test_pull_returns_result(self):
        parent = self.make_branch_and_tree('parent')
        p1 = parent.commit('1st post')
        mine = parent.controldir.sprout('mine').open_workingtree()
        m1 = mine.commit('my change')
        result = parent.branch.pull(mine.branch)
        self.assertIsNot(None, result)
        self.assertIs(mine.branch, result.source_branch)
        self.assertIs(parent.branch, result.target_branch)
        self.assertIs(parent.branch, result.master_branch)
        self.assertIs(None, result.local_branch)
        self.assertEqual(1, result.old_revno)
        self.assertEqual(p1, result.old_revid)
        self.assertEqual(2, result.new_revno)
        self.assertEqual(m1, result.new_revid)
        self.assertEqual([], list(result.tag_conflicts))

    def test_pull_overwrite(self):
        tree_a = self.make_branch_and_tree('tree_a')
        tree_a.commit('message 1')
        tree_b = tree_a.controldir.sprout('tree_b').open_workingtree()
        rev2a = tree_a.commit('message 2a')
        rev2b = tree_b.commit('message 2b')
        self.assertRaises(errors.DivergedBranches, tree_a.pull, tree_b.branch)
        self.assertRaises(errors.DivergedBranches,
                          tree_a.branch.pull, tree_b.branch,
                          overwrite=False, stop_revision=rev2b)
        # It should not have updated the branch tip, but it should have fetched
        # the revision if the repository supports "invisible" revisions
        self.assertEqual(rev2a, tree_a.branch.last_revision())
        if tree_a.branch.repository._format.supports_unreferenced_revisions:
            self.assertTrue(tree_a.branch.repository.has_revision(rev2b))
        tree_a.branch.pull(tree_b.branch, overwrite=True,
                           stop_revision=rev2b)
        self.assertEqual(rev2b, tree_a.branch.last_revision())
        self.assertEqual(tree_b.branch.last_revision(),
                         tree_a.branch.last_revision())

    def test_pull_overwrite_set(self):
        tree_a = self.make_branch_and_tree('tree_a')
        rev1 = tree_a.commit('message 1')

        tree_b = tree_a.controldir.sprout('tree_b').open_workingtree()
        rev2a = tree_a.commit('message 2a')
        rev2b = tree_b.commit('message 2b')
        self.assertRaises(errors.DivergedBranches, tree_a.pull, tree_b.branch)
        self.assertRaises(errors.DivergedBranches,
                          tree_a.branch.pull, tree_b.branch,
                          overwrite=set(), stop_revision=rev2b)
        # It should not have updated the branch tip, but it should have fetched
        # the revision if the repository supports "invisible" revisions
        self.assertEqual(rev2a, tree_a.branch.last_revision())
        if tree_a.branch.repository._format.supports_unreferenced_revisions:
            self.assertTrue(tree_a.branch.repository.has_revision(rev2b))
        tree_a.branch.pull(tree_b.branch, overwrite={'history'},
                           stop_revision=rev2b)
        self.assertEqual(rev2b, tree_a.branch.last_revision())
        self.assertEqual(tree_b.branch.last_revision(),
                         tree_a.branch.last_revision())
        tree_a.branch.pull(tree_b.branch, overwrite={'history', 'tags'},
                           stop_revision=rev2b)

    def test_pull_overwrite_set_tags(self):
        tree_a = self.make_branch_and_tree('tree_a')
        if not tree_a.branch.supports_tags():
            raise TestNotApplicable("branch does not support tags")
        rev1 = tree_a.commit('message 1')
        tree_a.branch.tags.set_tag('tag1', rev1)

        tree_b = tree_a.controldir.sprout('tree_b').open_workingtree()
        rev2b = tree_b.commit('message 2b')
        tree_b.branch.tags.set_tag('tag1', rev2b)
        rev1b = tree_a.commit('message 1b')
        tree_a.branch.get_config_stack().set('branch.fetch_tags', True)
        self.assertRaises(errors.DivergedBranches, tree_a.pull, tree_b.branch)
        self.assertRaises(errors.DivergedBranches,
                          tree_a.branch.pull, tree_b.branch,
                          overwrite=set(), stop_revision=rev2b)
        # It should not have updated the branch tip, but it should have fetched
        # the revision if the repository supports "invisible" revisions
        self.assertEqual(rev1b, tree_a.branch.last_revision())
        # It also should not have updated the tags
        self.assertEqual(tree_a.branch.tags.get_tag_dict(), {'tag1': rev1})
        if tree_a.branch.repository._format.supports_unreferenced_revisions:
            self.assertTrue(tree_a.branch.repository.has_revision(rev2b))
        tree_a.branch.pull(tree_b.branch, overwrite={'history'},
                           stop_revision=rev2b)
        self.assertEqual(rev2b, tree_a.branch.last_revision())
        self.assertEqual(tree_b.branch.last_revision(),
                         tree_a.branch.last_revision())
        self.assertEqual(rev1, tree_a.branch.tags.lookup_tag('tag1'))
        tree_a.branch.pull(tree_b.branch, overwrite={'history', 'tags'},
                           stop_revision=rev2b)
        self.assertEqual(rev2b, tree_a.branch.tags.lookup_tag('tag1'))

    def test_pull_merges_and_fetches_tags(self):
        """Tags are updated by br.pull(source), and revisions named in those
        tags are fetched.
        """
        # Make a source, sprout a target off it
        try:
            builder = self.make_branch_builder('source')
        except errors.UninitializableFormat:
            raise TestNotApplicable('uninitializeable format')
        source, rev1, rev2 = fixtures.build_branch_with_non_ancestral_rev(
            builder)
        target = source.controldir.sprout('target').open_branch()
        # Add a tag to the source, then pull from source
        try:
            source.tags.set_tag('tag-a', rev2)
        except errors.TagsNotSupported:
            raise TestNotApplicable('format does not support tags.')
        source.tags.set_tag('tag-a', rev2)
        source.get_config_stack().set('branch.fetch_tags', True)
        target.pull(source)
        # The tag is present, and so is its revision.
        self.assertEqual(rev2, target.tags.lookup_tag('tag-a'))
        target.repository.get_revision(rev2)

    def test_pull_stop_revision_merges_and_fetches_tags(self):
        """br.pull(source, stop_revision=REV) updates and fetches tags."""
        # Make a source, sprout a target off it
        try:
            builder = self.make_branch_builder('source')
        except errors.UninitializableFormat:
            raise TestNotApplicable('uninitializeable format')
        source, rev1, rev2 = fixtures.build_branch_with_non_ancestral_rev(
            builder)
        target = source.controldir.sprout('target').open_branch()
        # Add a new commit to the ancestry
        rev_2_again = builder.build_commit(message="Rev 2 again")
        # Add a tag to the source, then pull rev_2_again from source
        try:
            source.tags.set_tag('tag-a', rev2)
        except errors.TagsNotSupported:
            raise TestNotApplicable('format does not support tags.')
        source.get_config_stack().set('branch.fetch_tags', True)
        target.pull(source, stop_revision=rev_2_again)
        # The tag is present, and so is its revision.
        self.assertEqual(rev2, target.tags.lookup_tag('tag-a'))
        target.repository.get_revision(rev2)


class TestPullHook(per_branch.TestCaseWithBranch):

    def setUp(self):
        self.hook_calls = []
        super().setUp()

    def capture_post_pull_hook(self, result):
        """Capture post pull hook calls to self.hook_calls.

        The call is logged, as is some state of the two branches.
        """
        if result.local_branch:
            local_locked = result.local_branch.is_locked()
            local_base = result.local_branch.base
        else:
            local_locked = None
            local_base = None
        self.hook_calls.append(
            ('post_pull', result.source_branch, local_base,
             result.master_branch.base, result.old_revno,
             result.old_revid,
             result.new_revno, result.new_revid,
             result.source_branch.is_locked(), local_locked,
             result.master_branch.is_locked()))

    def test_post_pull_empty_history(self):
        target = self.make_branch('target')
        source = self.make_branch('source')
        branch.Branch.hooks.install_named_hook(
            'post_pull', self.capture_post_pull_hook, None)
        target.pull(source)
        # with nothing there we should still get a notification, and
        # have both branches locked at the notification time.
        self.assertEqual([
            ('post_pull', source, None, target.base, 0, revision.NULL_REVISION,
             0, revision.NULL_REVISION, True, None, True)
            ],
            self.hook_calls)

    def test_post_pull_bound_branch(self):
        # pulling to a bound branch should pass in the master branch to the
        # hook, allowing the correct number of emails to be sent, while still
        # allowing hooks that want to modify the target to do so to both
        # instances.
        target = self.make_branch('target')
        local = self.make_branch('local')
        try:
            local.bind(target)
        except branch.BindingUnsupported:
            # We can't bind this format to itself- typically it is the local
            # branch that doesn't support binding.  As of May 2007
            # remotebranches can't be bound.  Let's instead make a new local
            # branch of the default type, which does allow binding.
            # See https://bugs.launchpad.net/bzr/+bug/112020
            local = controldir.ControlDir.create_branch_convenience('local2')
            try:
                local.bind(target)
            except branch.BindingUnsupported:
                raise TestNotApplicable(
                    'default format does not support binding')
        source = self.make_branch('source')
        branch.Branch.hooks.install_named_hook(
            'post_pull', self.capture_post_pull_hook, None)
        local.pull(source)
        # with nothing there we should still get a notification, and
        # have both branches locked at the notification time.
        self.assertEqual([
            ('post_pull', source, local.base, target.base, 0,
             revision.NULL_REVISION, 0, revision.NULL_REVISION,
             True, True, True)
            ],
            self.hook_calls)

    def test_post_pull_nonempty_history(self):
        target = self.make_branch_and_memory_tree('target')
        target.lock_write()
        target.add('')
        rev1 = target.commit('rev 1')
        target.unlock()
        sourcedir = target.controldir.clone(self.get_url('source'))
        source = sourcedir.open_branch().create_memorytree()
        rev2 = source.commit('rev 2')
        branch.Branch.hooks.install_named_hook(
            'post_pull', self.capture_post_pull_hook, None)
        target.branch.pull(source.branch)
        # with nothing there we should still get a notification, and
        # have both branches locked at the notification time.
        self.assertEqual([
            ('post_pull', source.branch, None, target.branch.base, 1, rev1,
             2, rev2, True, None, True)
            ],
            self.hook_calls)
